/****************************************************************************************[Solver.h]
Copyright (c) 2005-2010, Niklas Een, Niklas Sorensson

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
associated documentation files (the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge, publish, distribute,
sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT
OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**************************************************************************************************/

#ifndef Solver_h
#define Solver_h

#include "MiniSat.h"
#include "SatELite.h"
#include "glueminisat/core/Solver.h"

//=================================================================================================


class Solver {
    MiniSat ::Solver*    minisat;
    SatELite::Solver*    satelite;
    GlueMiniSat::Solver* glueminisat;

    BasicSolverStats     basicStats;
    bool                 dummyOK;

    GlueMiniSat::Lit lit2lit(const Lit lit) {
        return GlueMiniSat::mkLit(var(lit), sign(lit));
    }
    void vec2vec(const vec<Lit>& in, GlueMiniSat::vec<GlueMiniSat::Lit>& out) {
        for (int i=0; i < in.size(); i++)
            out.push(lit2lit(in[i]));
    }
    lbool lbool2lbool(GlueMiniSat::lbool v) const {
        switch (GlueMiniSat::toInt(v)) {
            case 0: return toLbool(+1);
            case 1: return toLbool(-1);
            case 2: return toLbool( 0);
        }
        assert(false);
        return l_Undef;
    }

public:
    bool& ok_ref() {
        if (minisat)     return minisat->ok;
        if (satelite)    return satelite->ok;
        if (glueminisat) return glueminisat->getOK();
        assert(false);
        return dummyOK;
    }
//    vec<int>&         assigns_ref() { return (minisat != NULL) ? minisat->assigns : (satelite != NULL) ? satelite->assigns : glueminisat->assigns; }
//    vec<Lit>&         trail_ref  () { return (minisat != NULL) ? minisat->trail   : (satelite != NULL) ? satelite->trail   : glueminisat->trail; }
//    BasicSolverStats& stats_ref  () { return (minisat != NULL) ? (BasicSolverStats&)minisat->stats : (BasicSolverStats&)satelite->stats  ; }
    BasicSolverStats& stats_ref  () {
        if (minisat ) return (BasicSolverStats&)minisat->stats;
        if (satelite) return (BasicSolverStats&)satelite->stats;
        if (glueminisat) return basicStats;
        assert(false);
        return basicStats;
    }

    int assigns(Var v) const {
        if (minisat     != NULL) return minisat->assigns[v];
        if (satelite    != NULL) return satelite->assigns[v];
        if (glueminisat != NULL) return lbool2lbool(glueminisat->getAssign(v)).toInt();
        assert(false);
        return 0;
    }
    int trail_size() {
        if (minisat     != NULL) return minisat->trail.size();
        if (satelite    != NULL) return satelite->trail.size();
        if (glueminisat != NULL) return glueminisat->getTrailSize();
        assert(false);
        return 0;
    }
    Lit trail(int index) {
        if (minisat     != NULL) return minisat->trail[index];
        if (satelite    != NULL) return satelite->trail[index];
        if (glueminisat != NULL) return Lit(GlueMiniSat::var(glueminisat->getTrail(index)), GlueMiniSat::sign(glueminisat->getTrail(index)));
        assert(false);
        return lit_Error;
    }
    void setVerbosity(int level) {
        if (minisat    ) minisat->verbosity = level;
        if (satelite   ) satelite->verbosity = level;
        if (glueminisat) glueminisat->verbosity = level;
    }

//    void setVerbosity(int level) {
//        if (minisat != NULL)
//            minisat->verbosity = level;
//        else
//            satelite->verbosity = level; }

//    Var         newVar         (bool dvar = true)   { return (minisat != NULL) ? minisat->newVar(dvar) : satelite->newVar(dvar); }
//    bool        addClause      (const vec<Lit>& ps) { return (minisat != NULL) ? minisat->addClause(ps) : (satelite->addClause(ps), satelite->okay()); }
//    bool        addUnit        (Lit p)              { return (minisat != NULL) ? minisat->addUnit(p) : (satelite->addUnit(p), satelite->okay()); }
//    void        freeze         (Var x)              { if (minisat == NULL) satelite->freeze(x); }
//    void        suggestPolarity(Var x, lbool value) { if (minisat != NULL) minisat->polarity_sug[x] = toInt(value); else satelite->polarity_sug[x] = toInt(value); }
//    bool        solve     (const vec<Lit>& assumps) { return (minisat != NULL) ? minisat->solve(assumps) : satelite->solve(assumps); }
//    bool        solve          ()                   { vec<Lit> tmp; return solve(tmp); }
//    vec<lbool>& model          ()                   { return (minisat != NULL) ? minisat->model : satelite->model; }
//    bool        varElimed      (Var x)              { return (minisat != NULL) ? false : satelite->var_elimed[x]; }
//    bool        okay           ()                   { return (minisat != NULL) ? minisat->okay() : satelite->okay(); }
//    int         nVars          ()        const      { return (minisat != NULL) ? minisat->nVars() : satelite ? satelite->nVars() : glueminisat->nVars(); }


    Var newVar(bool dvar = true)   {
      //      printf("m %d, s %d, g %d\n", minisat, satelite, glueminisat);
        if (minisat)     return minisat    ->newVar(dvar);
        if (satelite)    return satelite   ->newVar(dvar);
        if (glueminisat) return glueminisat->newVar(true, dvar);
        assert(false);
        return 0;
    }
    bool addClause(const vec<Lit>& ps) {
        if (minisat)     return minisat    ->addClause(ps);
        if (satelite)    return (satelite->addClause(ps), satelite->okay());
        if (glueminisat) {
            GlueMiniSat::vec<GlueMiniSat::Lit> ps_;
            vec2vec(ps, ps_);
            return glueminisat->addClause_(ps_);
        }
        assert(false);
        return false;
    }
    bool addUnit(Lit p) {
        if (minisat)     return minisat  ->addUnit(p);
        if (satelite)    return (satelite->addUnit(p), satelite->okay());
        if (glueminisat) return glueminisat->addClause(lit2lit(p));
        assert(false);
        return false;
    }
    void freeze(Var x) {
        if (satelite)
            satelite->freeze(x);
	// else  assert(false);
    }
    void suggestPolarity(Var x, lbool value) {
        if (minisat ) minisat->polarity_sug[x] = toInt(value);
        else if (satelite) satelite->polarity_sug[x] = toInt(value);
	//  else assert(false);
    }
    bool solve(const vec<Lit>& assumps) {
        if (minisat)  return minisat->solve(assumps);
        if (satelite) return satelite->solve(assumps);
        if (glueminisat) {
            GlueMiniSat::vec<GlueMiniSat::Lit> assumps_;
            vec2vec(assumps, assumps_);
            bool ret = glueminisat->solve(assumps_);
            basicStats.conflicts    += glueminisat->conflicts;
            basicStats.decisions    += glueminisat->decisions;
            basicStats.propagations += glueminisat->propagations;
            basicStats.starts       += glueminisat->starts;
            return ret;
        }
        assert(false);
        return false;
    }
    bool solve() {
        vec<Lit> tmp; return solve(tmp);
    }
//    vec<lbool>& model          ()                   { return (minisat != NULL) ? minisat->model : satelite->model; }
    lbool model(Var v) {
        if (minisat)     return minisat ->model[v];
        if (satelite)    return satelite->model[v];
        if (glueminisat) return lbool2lbool(glueminisat->model[v]);
        assert(false);
        return l_Undef;
    }
    bool varElimed(Var x) {
        if (satelite)
            return satelite->var_elimed[x];
        else
            return false;
    }
    int ucore_vars(int idx) {
      if (glueminisat)
	return glueminisat->ucore_vars[idx];
//	return glueminisat->getUcore_vars(idx);
      else
        assert(false);
      return(0);
    }
    int ucore_varsSize() {
      if (glueminisat)
	return glueminisat->ucore_vars.size();
//	return glueminisat->getUcore_varsSize();
      else
        assert(false);
      return(0);
    }
    int ucore_clauses(int idx) {
      if (glueminisat)
	return glueminisat->ucore_clauses[idx];
//	return glueminisat->getUcore_clauses(idx);
      else
        assert(false);
      return(0);
    }
    int ucore_clausesSize() {
      if (glueminisat)
	return glueminisat->ucore_clauses.size();
//	return glueminisat->getUcore_clausesSize();
      else
        assert(false);
      return(0);
    }
    bool okay() {
        if (minisat)     return minisat->okay();
        if (satelite)    return satelite->okay();
        if (glueminisat) return glueminisat->okay();
        assert(false);
        return false;
    }
    int nVars() const {
        return (minisat != NULL) ? minisat->nVars() : satelite ? satelite->nVars() : glueminisat->nVars();
    }
    void exportCnf(cchar* filename) {
        if (minisat) {
            // Prohibited minisat-simplification before exporting CNF's
            // because of BUG that produce satisfiable CNF
            // even if it is un-satisfiable.
            //	    minisat->simplifyDB();
            //            reportf("ok=%s\n", minisat->ok ? "true" : "false");
            minisat->exportClauses(filename);
        } else if (satelite) {
            SatELite::opt_pre_sat = true;
            SatELite::output_file = filename;
            if (opt_verbosity >= 1) reportf("=================================[SATELITE+]==================================\n");
            satelite->simplifyDB(true);
        } else if (glueminisat) {
            glueminisat->toDimacs(filename);
        } else
            assert(false);
    }
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    //Solver(bool use_minisat) : minisat(use_minisat ? new MiniSat::Solver : NULL), satelite(use_minisat ? NULL : new SatELite::Solver) {}
    Solver(SolverT type) :
        minisat    (type == st_MiniSat  ? new MiniSat::Solver     : NULL),
        satelite   (type == st_SatELite ? new SatELite::Solver    : NULL),
        glueminisat(type == st_GlueMiniSat || type == st_MiniSat22 || type == st_Glucose10 ? new GlueMiniSat::Solver : NULL)
    {
      if (opt_extract_ucore) glueminisat->setUcoreExtract(opt_ucore_mode);
        if (type == st_MiniSat22) glueminisat->setMiniSat22Params();
        if (type == st_Glucose10) glueminisat->setGlucose10Params();
    }
};


//=================================================================================================
#endif
