(* Copyright 2018 Fabian Mitterwallner
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 *
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

open Util;;
open Rewritingx;;

module M = Monad;;
module Subst = Substitution;;
module H = Hashtbl;;
module Trans = Util.Monad.Transformer;;

let (>>=) = M.(>>=);;
let (>>) = M.(>>);;

(* Util Function *)
let find_default def tbl key =
  try H.find tbl key
  with Not_found -> def
;;

(* Finds all functions in the signature *)
let find_funs : Function.t list M.t =
  M.get >>= fun signature ->
  let names = Signature.fun_names signature in
  let funs = List.map (fun s -> Signature.find_fun s signature) names
  in
  M.return funs
;;


(* Returns a fresh constant. As a side effect the signature is changed.
 *)
let fresh_const : Term.t M.t =
  M.fresh_fun >>= fun c ->
  M.add_ari c 0 >>
  M.return (Term.Fun (c,[]))
;;

let n_fresh_consts : int -> Term.t list M.t =
fun n ->
  M.replicate n fresh_const
;;

(* Custom state monad for flattening *)
module SM = Util.Monad.Transformer.State (Trs) (M);;

let (>>=) = SM.(>>=);;
let (>>) = SM.(>>);;

(* Adds a new rule to the TRS *)
let add_rule : Rule.t -> unit SM.t =
fun rule -> SM.update (Trs.add rule)
;;

(* Flattens the right hand sides of the TRS conserving UNC.
 * Introduces new rules.
 *)
let flatten_right : Trs.t -> Trs.t M.t =
fun trs ->
  let hashtbl = H.empty () in
  let rec flatten_term : Term.t -> Term.t SM.t =
  fun term ->
    match term with
      | Term.Var _ -> SM.return term
      | Term.Fun (_,[]) -> SM.return term
      | Term.Fun (f,sbtrms) ->
          SM.map flatten_term sbtrms >>= fun sbtrms ->
          let new_rhs = Term.Fun (f, sbtrms) in
          (try SM.return (H.find hashtbl new_rhs)
          with Not_found ->
            SM.lift fresh_const >>= fun new_const ->
            add_rule (Rule.of_terms new_const new_rhs) >>
            (H.add hashtbl new_rhs new_const;
            SM.return new_const)
          )
  in
  let flatten_rule : Trs.rule -> unit SM.t =
    fun rule ->
      let rhs = Rule.rhs rule in
      match rhs with
      | Term.Var _ -> add_rule rule
      | Term.Fun (f, []) -> add_rule rule
      | Term.Fun (f, args) ->
          SM.map flatten_term args >>= fun args ->
          let lhs = Rule.lhs rule in
          let rhs' = Term.Fun (f,args) in
          add_rule (Rule.of_terms lhs rhs')
  in
    SM.execute Trs.empty (SM.map flatten_rule (Trs.to_list trs))
;;

(* Flattens left hand sides of the TRS conserving UNC.
 * Introduces new rules.
 *)
let flatten_left : Trs.t -> Trs.t M.t =
fun trs ->
  let hashtbl = H.empty ()
  in
  let addLeftRules : bool -> Term.t -> Term.t SM.t =
  fun is_normal term ->
    SM.lift fresh_const >>= fun new_const ->
    add_rule (Rule.of_terms term new_const) >>
    (if is_normal
    then SM.return ()
    else add_rule (Rule.of_terms new_const new_const)) >>
    SM.return new_const
  in
  let rec flatten_term : Term.t -> Term.t SM.t =
  fun term ->
    match term with
    | Term.Var _ -> SM.return term
    | Term.Fun (_,[]) -> SM.return term
    | Term.Fun (f,sbtrms) ->
        SM.map flatten_term sbtrms >>= fun sbtrms ->
        let new_lhs = Term.Fun (f, sbtrms) in
        (try SM.return (H.find hashtbl new_lhs)
        with Not_found ->
          addLeftRules (Trs.is_normal_form term trs) new_lhs >>= fun new_const ->
          H.add hashtbl new_lhs new_const;
          SM.return new_const)
  in
  let flatten_rule : Trs.rule -> unit SM.t =
    fun rule ->
      let lhs = Rule.lhs rule in
      match lhs with
      | Term.Var _ -> add_rule rule
      | Term.Fun (f, []) -> add_rule rule
      | Term.Fun (f, args) ->
          SM.map flatten_term args >>= fun args' ->
          let rhs = Rule.rhs rule in
          let lhs' = Term.Fun (f, args') in
          add_rule (Rule.of_terms lhs' rhs)
  in
    SM.execute Trs.empty (SM.map flatten_rule (Trs.to_list trs))
;;

let (>>=) = M.(>>=);;

let flatten : Trs.t -> Trs.t M.t =
fun trs -> flatten_right trs >>= flatten_left
;;


(****** Automata construction ******)
(* module A = Automata;; *)
module A = Automata.Default (Rewritingx);;
module Aut = A.Automaton;;
module AI = A.Initial;;
module AM = A.Monad;;
module Lhs = A.Lhs;;
module Rhs = A.Rhs;;

let (>>=) = AM.(>>=);;
let (>>) = AM.(>>);;

type automaton = Aut.t;;


(* [singleton_automata term] is the automaton, which only accepts the term
 * [term]
 *)
let singleton_automata : A.Term.t -> automaton AM.t =
fun term ->
  AI.specific (const (AM.liftM List.singleton AM.fresh_state))
  (A.Path.fresh_reuse) [term]
;;

(* [convert_automaton term trs] accepts all ground terms over the signature of
 * [trs] and [term] which are convertible to [term]
 *)
let convert_automaton : A.Term.t -> Trs.t -> automaton AM.t =
fun term trs ->
  AM.lift (find_funs) >>= fun funs ->
  let trs' = Trs.union trs (Trs.invert trs) in
  singleton_automata term >>= fun a ->
  A.Reachability.predecessors funs trs' a >>= fun res ->
  AM.return res
;;

(* [normalform_automata fs trs] is an automata which accepts all ground normal forms
 * of [trs] over the signature [fs].
 *)
let normalform_automata : Function.t list -> Trs.t -> automaton AM.t =
fun funs trs ->
  AM.liftM (Aut.ec ~r:true) (AI.normal funs trs) >>= fun res ->
  AM.return res
;;


(* [lhs_ground_instances trs] returns a list of all ground instances of the left
 * hand sides of [trs] over the signature of [trs].
 *)
let lhs_ground_instances : Term.t list -> Trs.t -> Term.t list =
fun consts trs ->
  let rec ground_instances : Term.t -> Term.t list =
  fun term ->
    if Term.is_ground term
    then [term]
    else
      let var = List.hd (Term.vars term) in
      let substs = List.map (fun trm -> Subst.singleton var trm) consts in
      List.flat_map ground_instances
        (List.map (fun sb -> Subst.apply_term sb term) substs)
  in

  List.flat_map ground_instances (Trs.lhs trs)
;;

(* [is_singleton a] checks if the automaton [a] accepts only a single term.
 *)
let is_singleton : automaton -> bool =
fun automaton ->
  let count = H.empty () in
  let witness = H.empty () in

  let rec diff : Lhs.t -> Lhs.t -> bool =
  fun t1 t2 ->
    (Lhs.root t1 <> Lhs.root t2) ||
    List.exists (fun (q1, q2) ->
          try diff (H.find witness q1) (H.find witness q2)
          with Not_found -> true
        ) (List.zip (Lhs.states t1) (Lhs.states t2))
  in

  let rules = Aut.trans automaton in
  let propagate_count () =
    List.foldl (fun changed (l, r) ->
      let r_states = Rhs.states r in
      let lh_state = Lhs.states l in

      List.foldl (fun changed r ->
        let counts = List.map (fun q -> find_default 0 count q) lh_state in
        match (find_default 0 count r) with
        | 0 when List.for_all ((==) 1) counts ->
            H.add count r 1;
            H.add witness r l;
            true
        | 1 when List.for_all ((==) 1) counts &&
                 try diff (H.find witness r) l
                 with Not_found -> true
                 ->
            H.add count r 2;
            true
        | c when List.for_all ((<=) 1) counts &&
                 List.exists ((==) 2) counts && c < 2 ->
            H.add count r 2;
            true
        | _ -> changed
      ) changed r_states

    ) false rules
  in
  let rec fixpoint () =
    if propagate_count ()
    then fixpoint ()
    else ()
  in

  fixpoint ();
  let final_states = Aut.finals automaton in
  let final_counts = List.map (find_default 0 count) final_states in

  if List.for_all ((==) 0) final_counts
  then false
  else if List.exists ((==) 2) final_counts
  then false
  else
    let reachable_finals =
      (List.filter (fun q -> find_default 0 count q == 1) final_states)
    in
    match reachable_finals with
    | [] -> false
    | (_::xs) ->
      List.foldl (fun acc (q1, q2) -> acc &&
        not (diff (H.find witness q1) (H.find witness q2))
      ) true (List.zip reachable_finals xs)
;;

(* [procedure trs] checks the UNC property for a given linear shallow [trs].
 *)
let procedure : Trs.t -> bool AM.t =
fun trs ->
  AM.lift (flatten trs) >>= fun trs ->
  AM.lift (n_fresh_consts 2) >>= fun new_consts ->
  let funs = Trs.funs trs @ (List.flat_map Term.funs new_consts) in
  normalform_automata funs trs >>= fun nf_autom ->
  let trs_consts = List.map (fun c -> Term.Fun (c,[])) (Trs.cons trs) in
  let grnd_instances =
    List.map A.Term.of_term (lhs_ground_instances (trs_consts @ new_consts) trs)
  in
  AM.for_all (fun term ->
    convert_automaton term trs >>= fun conv_autom ->
    Aut.inter nf_autom conv_autom >>= fun final_autom ->
    Aut.is_empty final_autom >>= fun empty_lang ->
    AM.return (is_singleton final_autom) >>= fun singleton_lang ->
    AM.return (empty_lang || singleton_lang)
  ) grnd_instances
;;

(* [is_unc] wraps [procedure] for running in the Rewritingx.Monad monad
 *)
let is_unc : Trs.t -> bool M.t =
fun trs ->
  let (>>=) = M.(>>=) in
  M.get >>= fun s ->
  let (r, s) = Either.right (AM.eval (s, A.Status.init) (procedure trs)) in
  (* M.set (fst s) >>= fun _ -> *) M.return r
;;
