(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Util;;
open Rewritingx;;

(*** MODULES ******************************************************************)
module F   = Format;;
module LL  = LazyList;;
module Pos = Position;;
module Sub = Substitution;;
module Sig = Signature;;

module Stream = Stream.Make (Sig);;

(*** TYPES ********************************************************************)
type flags = {
  help : bool ref;
  bwd : bool ref;
  fwd : bool ref;
  size : int ref;
  length : int ref;
  vpos : bool ref;
  innermost : bool ref;
};;

type t = Problem.t * Loop.t;;

(*** GLOBALS ******************************************************************)
let code = "unfold";;
let name = "Unfolding Processor";;
let comment = "This processor searches for loops by using unfoldings.";;
let keywords = ["unfolding"; "loop"; "nontermination"];;

let flags = {
  help = ref false;
  bwd = ref false;
  fwd = ref false;
  size = ref max_int;
  length = ref max_int;
  vpos = ref false;
  innermost = ref false;
};;

let spec =
  let spec = [
    ("-bwd", Arg.Set flags.bwd, "Use backward unfoldings to find loops.");
    ("-fwd", Arg.Set flags.fwd, "Use forward unfoldings to find loops.");
    ("--help", Arg.Set flags.help, "Prints information about flags.");
    ("-help", Arg.Set flags.help, "Prints information about flags.");
    ("-h", Arg.Set flags.help, "Prints information about flags.");
    ("-size", Arg.Set_int flags.size, "Maximal size of terms considered.");
    ("-length", Arg.Set_int flags.length, "Maximal length of loops considered.");
    ("-vpos", Arg.Set flags.vpos, "Rewrite also at variable positions.");
    ("-innermost", Arg.Set flags.innermost, "Only consider innermost loops.")]
  in
  Arg.alignx 80 spec
;;

let help = (comment, keywords, List.map Triple.drop_snd spec);;

(*** EXCEPTIONS ***************************************************************)
exception Too_large;;

(*** FUNCTIONS ****************************************************************)
let init _ = 
  flags.help := false;
  flags.bwd := false;
  flags.fwd := false;
  flags.length := max_int;
  flags.size := max_int;
  flags.vpos := false;
;;

(* Destructors *)
let make = Pair.make;;
let get_loop = snd;;
let get_ip = fst;;

(* Processor *)
(* rename all variables of a rule to fresh ones *)
let rename_rule rule =
  let rec renaming t ((sub, sign) as s) =
    match t with
    | Term.Var x ->
      if Sub.mem x sub then s
      else Sig.fresh_var sign |>> (fun y -> Sub.add x (Term.Var y) sub)
    | Term.Fun (f, ts) -> List.fold renaming ts s
  in
  renaming (Rule.lhs rule) @.@ Pair.make Sub.empty
  $>> (fun ren -> Rule.map (Sub.apply_term ren) rule)
;;

let unify size p st rule =
  let (s, t) = Rule.to_terms st in
  if Term.size s > size || Term.size t > size then raise Too_large
  else
    rename_rule rule
    $>> (fun (l', r') ->
          let mgu = Elogic.unify (Term.subterm p t) l' in
          Rule.of_terms (Sub.apply_term mgu s)
                        (Sub.apply_term mgu (Term.replace p r' t)))
;;

(* check whether [rule] is applicable at [p] in [t].
   if this is the case [rs] is extended appropriately. *)
let contractum size p (st, rs) rule sign =
  try
    unify size p st rule sign
    |>> (fun rule' -> (rule', rs @ [(rule, p)]))
    |> Option.some
  with Elogic.Not_unifiable | Too_large -> None
;;

let reducts_pos size trs strs p = Stream.of_values (contractum size p strs) trs;;

let reducts size trs pos strs = Stream.of_streams (reducts_pos size trs strs) pos;;

let forward size trs pos =
  Stream.flat_map (fun ((st, _) as strs) ->
  reducts size trs (pos (Rule.rhs st)) strs);;

let extractum size p (st, rs) rule sign =
  try
    unify size p (Rule.invert st) (Rule.invert rule) sign
    |>> (fun rule' -> (Rule.invert rule', (rule, p) :: rs))
    |> Option.some
  with Elogic.Not_unifiable | Too_large -> None
;;

let ancestors_pos size trs strs p = Stream.of_values (extractum size p strs) trs;;

let ancestors size trs pos strs = Stream.of_streams (ancestors_pos size trs strs) pos;;

let backward size trs pos = Stream.flat_map (fun ((st, _) as strs) ->
  ancestors size trs (pos (Rule.lhs st)) strs);;

let interleave size trs pos rules =
  Stream.merge (forward size trs pos rules) (backward size trs pos rules);;

let check (rule, rs) =
  let (l, r) = Rule.to_terms rule in
  Stream.of_values (fun u sign -> try
      ignore (Elogic.match_term u l);
      Some ((l, rs), sign)
    with Elogic.Not_matchable -> try
      let mgu = Elogic.unify l u in
      Some ((Sub.apply_term mgu l, rs), sign)
    with Elogic.Not_unifiable -> None) (Term.subterms r)
;;

let generate_loops n start step pos =
  Trs.to_list start
  |> Stream.of_values (fun rule s -> Some ((rule, [(rule, Pos.root)]), s))
  |> Stream.iterates (fun x ->
       Stream.sieve (fun (t, _) (u, _) -> not (Rule.subsumes u t)) (step pos x))
  |> Stream.take n (*only generate loops of length up to "n"*)
  |> Stream.concat
  |> Stream.flat_map check
  |> Stream.eval
;;

(* wrap into state monad *)
module M = Rewritingx.Monad;;
let (>>=) = M.(>>=);;
let (>>)  = M.(>>);;

(* select relative loops for relative problems *)
let select p ((t, rs), _) = 
 if Problem.is_rp p then 
  let rs = Trs.of_list (List.map fst rs) in
  not (Trs.is_empty (Trs.intersect rs (Problem.get_strict p)))
 else true
;;

let solve fs p =
  let configurate s = F.printf "%s@\n%!" s; flags.help := true in
  (try init (); Arg.parsex code spec fs with Arg.Bad s -> configurate s);
  if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
  if Problem.is_sp p || Problem.is_dp p || Problem.is_rp p then
    let (s, w) = Problem.get_sw p in
    (*FIXME: for relative termination, check that at least one strict rule is
    used in the loop*)
    let sw = Trs.union s w in
    let pos = if !(flags.vpos) then Term.pos else Term.funs_pos in
    let step = match (!(flags.fwd), !(flags.bwd)) with
      | (true,false)  -> forward
      | (false,true)  -> backward
      | (true,true)   -> interleave
      | (false,false) ->
        if not (Trs.is_duplicating sw) then forward
        else if not (Trs.is_duplicating(Trs.invert sw)) then backward
        else interleave
    in
    let size = !(flags.size) in
    let length = max !(flags.length) 1 in
    M.get >>= fun sign ->
    try
      let s = if Problem.is_dp p then s else sw in
      let loops =
        generate_loops length s (step size (Trs.to_list sw)) pos sign
        |> LL.filter (select p) (*may lead to nontermination if not satisfied by any element*)
        |> (LL.concat @.@ (LL.map (fun ((t, rs), st) ->
             LL.map (fun l -> (l, st)) (Loop.of_term_rules_positions t rs))))
        |> (if !(flags.innermost) || Problem.is_it p
          then LL.filter (Loop.is_innermost p @.@ fst)
          else id)
      in
      let (loop, sign) = LL.nth 0 loops in
      M.set sign >> M.return (Some (make p loop))
    with Failure e -> M.return None
  else M.return None
;;

(* Compare Functions *)
let equal p q = Problem.equal (get_ip p) (get_ip q);;

(* Printers *)
let fprintf fs fmt p =
  F.fprintf fmt "@[<1>%s:@\n" name;
  Loop.fprintf fmt (get_loop p) >>= fun _ ->
  F.fprintf fmt "@\n";
  List.hd fs fmt >>= fun _ ->
  F.fprintf fmt "@]";
  M.return ()
;;

let fprintfx _ p = Loop.fprintfx (get_loop p) (get_ip p);;

