(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Util;;
open Rewritingx;;

(*** MODULES ******************************************************************)
module C = Complexity;;
module F = Format;;
module M = Monad;;
module P = Problem;;
module XO = XmlOutput;;

(*** TYPES ********************************************************************)
type flags = {fresh : bool ref; help : bool ref};;
type t = P.t * (P.t list);;

(*** GLOBALS ******************************************************************)
let code = "acdp";;
let name = "ACDP Processor";;
let comment = "Transforms the given TRS into a AC-DP problem.";;
let keywords = ["acdp transformation";"actermination"];;
let flags = {fresh = ref true; help = ref false};;

let spec =
 let spec = [
  ("--help",Arg.Set flags.help,"Prints information about flags.");
  ("-help",Arg.Set flags.help,"Prints information about flags.");
  ("-h",Arg.Set flags.help,"Prints information about flags.");
  ("-sig",Arg.Clear flags.fresh,"Uses the original symbols as DP symbols.")]
 in
 Arg.alignx 80 spec
;;

let help = (comment,keywords,List.map Triple.drop_snd spec);;

(*** FUNCTIONS ****************************************************************)
let (>>=) = M.(>>=);;
let init _ = flags.fresh := true; flags.help := false;;

(* Destructors *)
let get_ip = fst;;
let get_ops = snd;;

(* Processor *)
let generate f trs =
 let ds = Trs.def_symbols trs in
 let generate r m =
  f (Rule.lhs r) >>= fun l ->
  let r = Rule.rhs r in
  let add m r = match Term.root r with
   | Some g when List.mem g ds && not (Term.is_proper_subterm r l) ->
     f r >>= fun r -> m >>= M.return @.@ Trs.add (Rule.of_terms l r)
   | _ -> m
  in
  List.foldl add m (Term.subterms r)
 in
 Trs.fold generate (M.return Trs.empty) trs >>= fun dps ->
 (M.return (P.create_dp (Trs.unique dps) trs P.Trivial))
;;

(*let is_ac f = Signature.is_theory Theory.AC -> Function.t -> t -> bool*)
(*let is_ac = Monad.liftM @.@ (Monad.is_theory Theory.AC)*)
 let is_ac f = 
  M.is_theory Theory.AC f >>= fun ac ->
  M.return ac
(*let is_ac f = true*)
;;

let generate_ac trs =
 let ds = Trs.def_symbols trs in
 
 let transform_rule rule accumulator=
    let l = Rule.lhs rule in
    let added = 
     match l with
      | Term.Fun (f,ts) -> is_ac f >>= fun b -> 
        if b then  M.fresh_var >>= fun x -> 
           let nlhs = Term.label_dp (Term.Fun(f, [l; Term.Var(x)])) in
           let nrhs = Term.label_dp (Term.Fun(f, [Rule.rhs rule; Term.Var(x)])) in
           nlhs >>= fun nlhs -> nrhs >>= fun nrhs ->
           let nrule = Rule.of_terms nlhs nrhs in
           accumulator >>= fun accumulator -> M.return (Trs.add nrule accumulator)
        else 
           accumulator
      | _ -> accumulator
     in
     added
 in
 
 let dependancy = Trs.fold transform_rule (M.return Trs.empty) trs in
 dependancy >>= fun dps -> (M.return (P.create_dp (Trs.unique dps) trs P.Trivial))
;;
 
let solve fs p =
Format.printf "toto1";Format.print_newline ();
 let configurate s = F.printf "%s@\n%!" s; flags.help := true in
 (try init (); Arg.parsex code spec fs with Arg.Bad s -> configurate s);
 if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
 (*if P.is_sp p & Trs.is_trs (P.get_trs p) then
  (if !(flags.fresh) then [generate Term.label_dp (P.get_trs p)]
  else generate M.return (P.get_trs p)) >>=
  (M.return @.@ Option.map (Pair.make p @.@ flip P.adapt p))
 else M.return*) 
 let dpr = generate Term.label_dp (P.get_trs p) in
 let dpac = generate_ac (P.get_trs p) in
 let res = dpr>>= fun dpprob -> dpac >>= fun dpacf -> match P.is_empty (P.adapt dpprob p), P.is_empty (P.adapt dpacf p) with
                                                      | true, true -> M.return None
                                                      | true, false -> M.return (Some(p, [P.adapt dpacf p]))
                                                      | false, true -> M.return (Some(p, [P.adapt dpprob p]))
                                                      | false, false ->M.return (Some(p, [ P.adapt dpprob p;P.adapt dpacf p])) in
 (*let test = dpr>>= (fun x -> match x with 
                             | Some(y) -> M.return y
                             | _ -> failwith "error") in
                             
 let res = test>>= (fun dpprob -> M.return (Some(p, [P.adapt dpprob p;p;P.adapt dpprob p]))) in*)
 
 (*M.return (Some(p, (*M.sequence [dpr]*)[p]))*)
 res
;;

let is_empty p = List.for_all Problem.is_empty (get_ops p);;

(* Complexity Bounds *)
let complexity c _ = C.mul c C.other;;

(* Compare Functions *)
let equal p q = P.equal (get_ip p) (get_ip q);;

(* Printers
let fprintf fs fmt p =
 Format.printf "toto2";Format.print_newline ();
 Format.printf "%d %d\n" (List.length (get_ops p)) (List.length fs);
 F.fprintf fmt "@[<1>%s:@\n" name; 
 
 let test = List.fold_left (fun _ x -> F.fprintf fmt "\ntotostart\n"; let r= P.fprintfm fmt x in F.fprintf fmt "\ntotoend\n"; r) (M.return ()) (get_ops p) in
 
 
 test >>= fun _ ->
 F.fprintf fmt "@\n"; List.hd fs fmt >>= fun _ -> (List.nth fs 2) fmt >>= fun _ ->M.return (F.fprintf fmt "@]")
;;*)

let fprintf fs fmt p =
 (*Format.printf "%d %d\n" (List.length (get_ops p)) (List.length fs);*)
 F.fprintf fmt "@[<1>%s:@\n" name; 
 let fprintfi i fmt p =
  P.fprintfm fmt p >>= fun _ -> F.fprintf fmt "@\n"; List.nth fs i fmt
 in
 M.fprintfi fprintfi "@\n@\n" fmt (get_ops p) >>= fun _ ->
 M.return (F.fprintf fmt "@]")
;;

let fprintfx fs prob =
  if List.length (get_ops prob) = 0 then failwith "No problem to print" else 
  let op  = List.hd (get_ops prob) in
  let dps = P.get_dps op in
  XO.node "dpTrans" [
    P.fprintfx op;
    XO.bool "markedSymbols" true;
    if Trs.is_empty dps
      then XO.node "dpProof" [XO.leaf "pIsEmpty"]
      else List.hd fs
     
];;
