open Util;;
open Rewritingx;;

module C = Complexity;;
module F = Format;;
module G = Graph;;
module M = Monad;;
module P = Problem;;
module XO = XmlOutput;;

(*** TYPES ********************************************************************)
type flags = {
  help : bool ref;
  cmd : string ref;
  expected : string ref;
  full_match : bool ref; 
};;
type t = P.t * P.t;;

(*** GLOBALS ******************************************************************)
let code = "odg";;
let name = "Oracle Dependency Graph Approximation Processor";;
let keywords = ["dependency graph"; "approximation"; "termination"; "oracle"];;
let flags = {
  help = ref false;
  cmd = ref "";
  expected = ref "YES";
  full_match = ref false;
};;

let comment =
  "Remove all edges from the current DG according to an external oracle."
;;

let spec =
  let spec = [
    ("--help", Arg.Set flags.help, "Prints information about flags.");
    ("-help", Arg.Set flags.help, "Prints information about flags.");
    ("-h", Arg.Set flags.help, "Prints information about flags.");
    ("-cmd", Arg.Set_string flags.cmd,
      "Path to oracle executable for edge estimation.");
    ("-expect", Arg.Set_string flags.expected,
      "Expected output of successful oracle, default \"YES\"");
    ("-full", Arg.Set flags.full_match,
      "If false, match only first line of oracle output against expected \
      answer (-expect), otherwise match full output (default: false)");
  ] in
  Arg.alignx 80 spec
;;

let help = (comment,keywords, List.map Triple.drop_snd spec);;

(*** FUNCTIONS ****************************************************************)
let (>>=) = M.(>>=);;
let init _ =
  flags.help := false;
  flags.cmd := "";
  flags.expected := "YES";
  flags.full_match := false;
;;

(* Destructors *)
let get_ip = fst;;
let get_op = snd;;

(* Processor *)

(*add rules: eq(x,x) -> true, eq(t, u) -> false*)
let mk_extended trs t u =
  M.fresh_nary 0 >>= fun ttrue ->
  M.fresh_nary 0 >>= fun ffalse ->
  M.fresh_nary 2 >>= fun eq ->
  M.fresh_var    >>= fun x ->
  trs
  |> Trs.add (Rule.of_terms (Term.Fun (eq, [Term.Var x; Term.Var x])) (Term.Fun (ttrue, [])))
  |> Trs.add (Rule.of_terms (Term.Fun (eq, [t; u])) (Term.Fun (ffalse, [])))
  |> M.return
;;

let fpf_signature fmt fs =
  F.fprintf fmt "SIGNATURE @[";
  M.fprintf (fun fmt f ->
    M.fprintf_fun fmt f >>= fun _ ->
    M.find_ari f >>= fun a ->
    F.fprintf fmt ": %a -> ANY" (List.fprintf F.pp_print_string " ") (List.replicate a "ANY");
    M.return ()
  ) "@\n" fmt fs >>= fun _ ->
  M.return (F.fprintf fmt "@]@.")
;;

let fpf_ordering fmt fs =
  F.fprintf fmt "ORDERING @[LPO ";
  M.fprintf (M.fprintf_fun) " > " fmt fs >>= fun _ ->
  M.return (F.fprintf fmt "@]@.")
;;

let fpf_variables fmt xs =
  F.fprintf fmt "VARIABLES @[";
  M.fprintf (M.fprintf_var) ", " fmt xs >>= fun _ ->
  M.return (F.fprintf fmt " : ANY@]@.")
;;

let rec fpf_term fmt = function
  | Term.Var x -> M.fprintf_var fmt x
  | Term.Fun (f, ts) ->
    F.fprintf fmt "@["; M.fprintf_fun fmt f >>= fun _ ->
    M.find_ari f >>= fun a ->
    if a > 0 then
      (F.fprintf fmt "@[("; M.fprintf fpf_term "," fmt ts) >>= fun _ ->
      M.return (F.fprintf fmt ")@]@]")
    else
      (F.fprintf fmt "@["; M.fprintf fpf_term "," fmt ts) >>= fun _ ->
      M.return (F.fprintf fmt "@]@]")
;;

let fpf_eq fmt r =
  F.fprintf fmt "@["; fpf_term fmt (Rule.lhs r) >>= fun _ ->
  F.fprintf fmt "@ =@ "; fpf_term fmt (Rule.rhs r) >>= fun _ ->
  M.return (F.fprintf fmt "@]")
;;

let fpf_equations fmt trs =
  F.fprintf fmt "EQUATIONS @["; M.fprintf fpf_eq "@\n" fmt (Trs.to_list trs) >>= fun _ ->
  M.return (F.fprintf fmt "@]@.")

let fpf_conclusion fmt t u =
  F.fprintf fmt "CONCLUSION @["; fpf_eq fmt (Rule.of_terms t u) >>= fun _ ->
  M.return (F.fprintf fmt "@]@.")
;;

let fprintf_waldmeister fmt trs t u =
  F.fprintf fmt "NAME edge-oracle@.";
  F.fprintf fmt "MODE PROOF@.";
  F.fprintf fmt "SORTS ANY@.";
  let trs' = Trs.add (Rule.of_terms t u) trs in
  let fs = Trs.funs trs' in
  let xs = Trs.vars trs' in
  fpf_signature fmt fs >>= fun _ ->
  fpf_ordering fmt fs >>= fun _ ->
  fpf_variables fmt xs >>= fun _ ->
  fpf_equations fmt trs >>= fun _ ->
  fpf_conclusion fmt t u >>= fun _ ->
  M.return (F.fprintf fmt "%!")
;;

let oracle trs t u =
  if !(flags.cmd) <> "" then
    let (tempname, hdl) = Filename.open_temp_file "oracle." ".edge" in
    fprintf_waldmeister (Format.formatter_of_out_channel hdl) trs t u >>= fun _ ->
    close_out hdl;
    let hdl = Unix.open_process_in (!(flags.cmd) ^ " " ^ tempname ^ " 2>/dev/null") in
    let answer = ref true in
    let r = try
      if !(flags.full_match) then
        let r = Str.regexp !(flags.expected) in
        let continue = ref true in
        while !continue do
          let a = input_line hdl in
          try
            ignore (Str.search_forward r a 0);
            answer := false;
            continue := false
          with Not_found -> ()
        done;
        !answer
      else
        let a = input_line hdl in 
        if a = !(flags.expected) then false else true
    with End_of_file -> !answer in
    close_in hdl;
    Sys.remove tempname;
    M.return r
  else M.return true
;;

let rename_labeled s t =
  let rec rename m = function
    | Term.Var _ as x -> M.return (x, m)
    | Term.Fun (f, ts) ->
      let g ti (ts, m) = M.liftM (Util.Pair.apply (flip List.cons ts) id) (rename m ti) in
      M.foldr g ([], m) ts >>= fun (ts', m') ->
      M.is_labeled f >>= function
      | false -> M.return (Term.Fun (f, ts'), m')
      | true ->
        if List.mem_assoc f m' then M.return (Term.Fun (List.assoc f m, ts'), m')
        else
          M.find_ari f >>= M.fresh_nary >>= fun f' ->
          M.return (Term.Fun (f', ts'), (f, f') :: m')
  in
  rename [] s >>= fun (s', m) ->
  rename m t >>= fun (t', m') ->
  M.return (s', t', m')
;;

let is_arc trs (x, y) =
  let t = Rule.rhs x in
  let u = Rule.lhs y in
  Term.rename u >>=
  rename_labeled t >>= fun (t', u', _) ->
  oracle trs t' u'
;;

let filter_edges trs g =
 Graph.fold (fun n ms m -> m >>= fun g ->
   let add m ms c = Monad.return (if c then m::ms else ms) in
   Monad.foldl (fun ms m -> is_arc trs (n,m) >>= add m ms) [] ms >>=
   (Monad.return @.@ flip (Graph.add n) g)) g (Monad.return Graph.empty)
;;

let solve fs p =
  let configurate s = F.printf "%s@\n%!" s; flags.help := true in
  (try init (); Arg.parsex code spec fs with Arg.Bad s -> configurate s);
  if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
  if P.is_dp p then
    let trs = Problem.get_trs p and dps = Problem.get_dps p in
    let g = Problem.the_graph p in
    filter_edges trs (Graph.restrict (Trs.to_list dps) g) >>= fun h ->
    let progress = not (Graph.equal h g) in
    let q = Problem.set_dg (Problem.Partial h) dps trs p in
    Monad.return (if progress then Some (p, q) else None)
  else Monad.return None
;;

(* Complexity Bounds *)
let complexity c _ = C.mul c C.other;;

(* Compare Functions *)
let equal p q = P.equal (get_ip p) (get_ip q);;

(* Printers *)
let fprintf fs fmt p =
 F.fprintf fmt "@[<1>%s:@\n" name;
 P.fprintfm ~g:true fmt (get_op p) >>= fun _ ->
 F.fprintf fmt "@\n"; List.hd fs fmt >>= fun _ ->
 M.return (F.fprintf fmt "@]")
;;

let fprintfx fs p = XO.node "depGraphProc" [
  P.fprintfx ~g:true (get_op p);
  List.hd fs];;

