(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Util;;
open Rewritingx;;

(*** MODULES ******************************************************************)
module C = Complexity;;
module F = Format;;
module M = Monad;;
module P = Problem;;
module XO = XmlOutput;;

(*** TYPES ********************************************************************)
type flags = {
 help : bool ref;
 cmd : string ref;
 expect : string ref;
 prf : bool ref;
};;

type t = {
  p_ip : P.t;
  p_op : P.t;
  p_cmd : string;
  p_proof : string;
};;

(*** GLOBALS ******************************************************************)
let code = "oracle";;
let name = "Oracle processor";;
let comment = "Invoke external tool on current problem.";;
let keywords = ["external tool";"oracle"];;

let flags = {
 help = ref false;
 cmd = ref "";
 expect = ref "YES";
 prf = ref false;
};;

let spec =
 let spec = [
  ("--help",Arg.Set flags.help,"Prints information about flags.");
  ("-help",Arg.Set flags.help,"Prints information about flags.");
  ("-h",Arg.Set flags.help,"Prints information about flags.");
  ("-cmd",Arg.Set_string flags.cmd,"Path to oracle executable, e.g. \"./termtool\".");
  ("-expect",Arg.Set_string flags.expect,"Expected first line of output, default: \"YES\"");
  ("-p",Arg.Set flags.prf,"Print proof output by oracle executable.");
 ] in
 Arg.alignx 80 spec
;;

let help = (comment,keywords,List.map Triple.drop_snd spec);;

(*** FUNCTIONS ****************************************************************)
let (>>=) = M.(>>=);;
let init _ =
  flags.help := false;
  flags.cmd := "";
  flags.expect := "YES";
  flags.prf := false;
;;


(* Contructors *)
let make p prf = {
 p_ip = p;
 p_op = P.set_sw Trs.empty Trs.empty p;
 p_cmd = !(flags.cmd);
 p_proof = prf;
};;

(* Destructors *)
let get_ip p = p.p_ip;;
let get_op p = p.p_op;;

(* Processor *)

let fprintf_rule strict fmt r =
  F.fprintf fmt "  @["; Term.fprintfm fmt (Rule.lhs r) >>= fun _ ->
  F.fprintf fmt "@ %s@ " (if strict then "->" else "->=");
  Term.fprintfm fmt (Rule.rhs r) >>= fun _ ->
  F.fprintf fmt "@]@."; M.return ()
;;

let fprintf_problem fmt p =
  let (s,w) = P.get_sw p in
  let vs = List.unique_hash (Trs.vars s @ Trs.vars w) in
  let fs = List.unique_hash (Trs.funs s @ Trs.funs w) in
  M.filter (M.is_theory Theory.AC) fs >>= fun acfs ->
  F.fprintf fmt "(VAR@.  @[";
  M.iter (fun v -> M.fprintf_var fmt v >>=
    fun _ -> F.fprintf fmt "@ "; M.return ()) vs >>= fun _ ->
  F.fprintf fmt "@]@.)@.";
  (if not (List.is_empty acfs) then F.fprintf fmt "(THEORY@.(AC@.  @[" else ());
  M.iter (fun f -> M.fprintf_fun fmt f >>=
    fun _ -> F.fprintf fmt "@ "; M.return ()) acfs >>= fun _ ->
  (if not (List.is_empty acfs) then F.fprintf fmt "@]@.))@." else ());
  F.fprintf fmt "(RULES@.";
  M.iter (fun r -> fprintf_rule true  fmt r) (Trs.to_list s) >>= fun _ ->
  M.iter (fun r -> fprintf_rule false fmt r) (Trs.to_list w) >>= fun _ ->
  F.fprintf fmt ")@."; M.return ()
;;

let solve fs p =
  let configurate s = F.printf "%s@\n%!" s; flags.help := true in
  (try init (); Arg.parsex code spec fs with Arg.Bad s -> configurate s);
  if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
  let (tempname, hdl) = Filename.open_temp_file "csi." ".trs" in
  fprintf_problem (Format.formatter_of_out_channel hdl) p >>= fun _ ->
  close_out hdl;
  let hdl = Unix.open_process_in (!(flags.cmd) ^ " " ^ tempname ^ " 2>/dev/null") in
  let r = try
    let e = input_line hdl in
    let prf = if !(flags.prf) then read_channel hdl else "" in
    if e = !(flags.expect) then Some (make p prf) else None
   with End_of_file -> None in
  close_in hdl;
  Sys.remove tempname;
  M.return r
;;

(* Complexity Bounds *)
let complexity c _ = C.mul c C.other;;

(* Compare Functions *)
let equal p q = P.equal (get_ip p) (get_ip q);;

(* Printers *)
let fprintf fs fmt p =
 let prf = String.to_char_list p.p_proof in
 let prf = List.flat_map (fun c -> if c = '\n' then ['@';'\n'] else [c]) prf in
 let prf = Scanf.format_from_string (String.of_char_list prf) "" in
 F.fprintf fmt ("@[<1>%s (%s):@\n" ^^ prf ^^ "@\n") name p.p_cmd;
 List.hd fs fmt >>= fun _ -> M.return (F.fprintf fmt "@]")
;;

let fprintfx _ _ = XO.leaf "rMagic";;
