(* Copyright 2018 Jonas Schöpf
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 *
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Rewritingx;;
open Util;;

(*** MODULES *********************************************************)
module Fun = Function;;
module Parse = Parsec.MakeCharParser(Parsec.StringInput);;
module Sig = Signature;;

(*** FUNCTIONS ****************************************************************)
let is_fun signature f : bool =
  let sym = String.xml_entities_encode (String.of_char_list f) in
  try ignore (Sig.find_fun sym signature); true with Not_found -> false
;;

let to_fun signature f : Fun.t =
  try Sig.find_fun (String.xml_entities_encode f) signature with
    Not_found -> failwith ("function symbol '" ^ f ^ "' unknown")
;;

let to_arity signature f : int = try Sig.find_ari f signature with
   Not_found -> failwith ("function symbol '" ^ (Sig.find_fun_name f signature) ^ "' unknown")
;;

let (>>) = Parse.(>>);;
let (>>=) = Parse.(>>=);;
let (<?>) = Parse.(<?>);;
let (<|>) = Parse.(<|>);;

(*find longest prefix of valid characters that satisfies 'p'*)
let longest_till p : string Parse.t =
  let valid_char = Parse.noneof "(), \t\n\r" in
  (Parse.tempt @@ Parse.lookahead @@ Parse.many1 valid_char) >>= fun xs ->
  match List.prefixes xs |> List.rev |> List.find_option p with
    | None -> Parse.failwith (
      "no prefix of \"" ^ String.of_char_list xs ^ "\" is a valid function symbol")
    | Some rs ->
      Parse.count (List.length rs) Parse.any >>
      Parse.return (String.of_char_list rs)
;;

(*parse a single function symbol*)
let parse_fun signature : Fun.t Parse.t = Parse.lex (
  longest_till (is_fun signature) >>= (Parse.return @.@ to_fun signature)
) <?> "function symbol"
;;

type 'a prec_types = Gt of 'a | Eq of 'a | Ge of 'a | No of 'a

type precedence = (Fun.t * Fun.t) prec_types list;;

type 'a bool_structure =
  | ATM of 'a
  | NEG of 'a bool_structure
  | DIS of 'a bool_structure list
  | CON of 'a bool_structure list
;;

let rec exists_atom p = function
  | ATM x -> p x
  | NEG x -> exists_atom p x
  | DIS xs | CON xs -> List.exists (exists_atom p) xs
;;

let is_quasi_prec : precedence bool_structure -> bool = exists_atom (List.exists @@ function
  | Eq _ | Ge _ -> true
  | _ -> false)
;;

(*parse comma separated list*)
let commas ?(sep = ",") = Parse.sep_by1 (Parse.lex @@ Parse.string sep);;

let parse_boolean parse_atom : 'a bool_structure Parse.t =
  (*parse between parentheses*)
  let pars p = (*Note: type inference doesn't work with partial application*)
    Parse.between (Parse.lex @@ Parse.char '(') (Parse.lex @@ Parse.char ')') p
  in
  (*parse comma separated list between parentheses*)
  let args = pars @.@ commas in
  let rec pb pa =
    (Parse.tempt pa >>= fun a -> Parse.return (ATM a)) <|> (
      (Parse.string "AND" <|> Parse.string "OR" <|> Parse.string "NOT") >>= function
      | ['N';'O';'T'] -> pars (pb pa) >>= fun x -> Parse.return (NEG x)
      | ['A';'N';'D'] -> args (pb pa) >>= fun x -> Parse.return (CON x)
      | ['O';'R']     -> args (pb pa) >>= fun x -> Parse.return (DIS x)
    )
  in
  (Parse.tempt (commas parse_atom) >>= fun xs ->
    Parse.return (CON (List.map (fun x -> ATM x) xs))) <|>
  pb parse_atom
;;

let parse_prec signature : precedence Parse.t =
  let parse_gt_fun =
    Parse.lex ((Parse.tempt @@ Parse.string ">=") <|> Parse.string ">" <|> Parse.string "=") >>= fun op ->
    parse_fun signature >>= fun f ->
    match op with
     | ['>'] -> Parse.return (Gt f)
     | ['>'; '='] -> Parse.return (Ge f)
     | ['='] -> Parse.return (Eq f)
  in
  let rec create_prec_pairs = function
    | No x :: Gt y :: xs -> Gt (x, y) :: create_prec_pairs (No y :: xs)
    | No x :: Ge y :: xs -> Ge (x, y) :: create_prec_pairs (No y :: xs)
    | No x :: Eq y :: xs -> Eq (x, y) :: create_prec_pairs (No y :: xs)
    | _ -> []
  in
  parse_fun signature >>= fun f ->
  Parse.many1 parse_gt_fun >>=
  Parse.return @.@ create_prec_pairs @.@ cons (No f)
;;

let parse p s =
  let input = Parsec.StringInput.of_string s in
  match Parse.parse (Parse.between Parse.spaces Parse.eoi p) input with
  | Left e -> failwith (Parsec.Error.to_string e)
  | Right x -> Some x
;;

let parse_prec_template signature = function
  | "" -> None
  | s -> parse (parse_boolean @@ parse_prec signature) s
;;

let parse_nat : int Parse.t = (
  Parse.lex @@ Parse.many1 Parse.digit >>=
  Parse.return @.@ int_of_string @.@ String.of_char_list
) <?> "natural number"
;;

(* Given "constructor functions"
  mk_a  ... for arbitrary/unknown monomials
  mk_c  ... for constants
  mk_v  ... for variables
  mk_cv ... for coefficient times variable
and a parser
  parse_const ... for constants
parse and return a linear polynomial.*)
let parse_linpoly mk_a mk_c mk_v mk_cv parse_const =
  let parse_var = (
    Parse.many1 Parse.letter >> parse_nat >>= Parse.return @.@ Option.some
  ) <?> "variable"
  in
  let const_var =
    Parse.option None (parse_const >>= Parse.return @.@ Option.some) >>= fun c ->
    Parse.option None parse_var >>= fun i ->
    match (c, i) with
    | (Some c, Some i) -> Parse.return @@ mk_cv c i
    | (None, Some i) -> Parse.return @@ mk_v i
    | (Some c, None) -> Parse.return @@ mk_c c
    | _ -> Parse.failwith "expecting monomial"
  in
  let unk = (Parse.lex @@ Parse.char '_' >> Parse.return mk_a) <?> "underscore" in
  let mon = unk <|> const_var in
  let sum = commas ~sep:"+" mon in
  sum
;;

let parse_fun_eq signature : Fun.t Parse.t = Parse.tempt (
  parse_fun signature >>= fun f ->
  Parse.lex @@ Parse.char '=' >>
  Parse.return f
) <?> "function symbol followed by '='"
;;

let parse_inters s pi : 'a list Parse.t =
  Parse.lex @@ Parse.many1 (parse_fun_eq s) >>= fun fs ->
  pi >>= fun is ->
  Parse.return @@ List.map (fun f -> ((f, to_arity s f), is)) fs
;;

