(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 *
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** SUBMODULES ***************************************************************)
module Var = Variable;;
module Fun = Function;;
module M = Monad;;
module A = Array;;
module F = Format;;

(*** OPENS ********************************************************************)
open Prelude;;
open Util;;

(* An inefficient diophantine equation solver based on the completion procedure
   described (for instance) in Clausen and Fortenbacher: Efficient solution of
	 linear diophantine equations (1983). *)
module Dio = struct
	(* the unit vector with a 1 at position (i-1) *)
	let ei i n =
	  let a = A.make n 0 in
	  A.set a (i-1) 1;
	  A.to_list a
	;;

	let add xs ys = List.map (fun (x,y) -> x+y) (List.zip xs ys);;
	let add2 (a,b) (c,d) = (add a c, add b d);;
	let zero n = A.to_list (A.make n 0);;
	(* initial solution proposals:
	   fill unit vector for a positive variable up with zeros for the negatives *)
	let p0 m n = List.map (fun i -> ei i m, zero n) (List.range 1 (m+1));;
	let sum = List.fold_left2 (fun s a b -> s + a*b) 0;;

	(* the defect *)
	let d xs ys (px,py) = (sum xs px) - (sum ys py);;
	(* component-wise smaller or equal *)
	let leq = List.fold_left2 (fun v a b -> v && (a <= b)) true;;

  (* check whether proposal (px, py) is minimal wrt m, i.e., there is no smaller
	  solution in m *)
	let minimal (px, py) m =
    not (List.exists (fun (x,y) -> (leq x px) && (leq y py)) m)
	;;

  (* main completion procedure to compute complete list of minimal solutions:
	   - d is a defect function
		 - en and em are lists of unit vectors
		 - p is a list of solution proposals (initially supposed to be p0)
		 - m_all is the already computed list of minimal solutions
	*)
	let rec complete d en em p m_all =
	if p = [] then List.unique m_all
	else
	  (* partition nonminimal proposals into negative/positive defect *)
		let pn,pp =
			List.partition (fun p -> d p < 0) (List.filter (fun x -> d x <> 0) p)
		in
		(* try to correct proposals by adding unit vector *)
		let q' =
		  (List.flat_map (fun ej -> List.map (add2 ej) pn) en) @
		  (List.flat_map (fun ei -> List.map (add2 ei) pp) em)
		in
		(* those with defect 0 are new minimal ones *)
		let m' = List.filter (fun s -> d s = 0) q' in
		let p' = List.filter (fun p -> minimal p m_all) (List.diff q' m') in
		complete d en em p' (m' @ m_all)
	;;

  (* Compute basis of solutions for homogeneous diophantine equation
	   a_1 x_1 + ... + a_m x_m - b_1 y_1 - ... - b_n y_n = 0
	 *)
	let basis aa bb =
		let m,n = List.length aa, List.length bb in
		let d = d aa bb in (* defect function, hiding coefficients *)
		let em = List.map (fun i -> ei i m, zero n) (List.range 1 (m+1)) in
		let en = List.map (fun i -> zero m, ei i n) (List.range 1 (n+1)) in
		complete d em en (p0 m n) []
	;;

  (* Compute basis of solutions for homogeneous diophantine equation
	   x_i_1 + ... + x_i_k - y_j_1 - ... - y_j_l = 0
		 (without coefficients)
	 *)
	let solve xs ys =
		let c s = List.map (fun g -> List.hd g, List.length g) (List.group s) in
		let xs', ys' = c xs, c ys in
		let bs = basis (List.map snd xs') (List.map snd ys') in
		let vars = List.map fst (xs'@ys') in
	  List.map (fun (xb,yb) -> List.zip vars (xb@yb)) bs
	;;
end


(*** MODULES ******************************************************************)
module Make (L : LABEL) = struct
 (*** MODULES *****************************************************************)
 module M = M.Make (L);;
 module S = Substitution.Make (L);;
 module Sig = Signature.Make (L);;
 module T = Term.Make (L);;
 module Rule = Rule.Make(L);;
 module Sub = Substitution.Make(L);;
 module VSub = Replacement.Make(Var)(Var);;
 module TSub = Replacement.Make(T)(T);;

 (*** TYPES *******************************************************************)
 type 'a m = 'a M.t;;
 type substitution = S.t;;
 type term = T.t = Var of Var.t | Fun of Fun.t * term list;;

 (*** OPENS (2) ***************************************************************)
 open M;;

 (*** FUNCTIONS ***************************************************************)
 (* zero substitution list {{}} *)
 let zero_list = [Sub.empty]

 let fresh_var =
	fresh_var >>= fun v ->
	create_var_name v >>
	return v
 ;;

 let is_ac_symbol = M.is_theory Theory.AC

 let print_sub sub =
  let sub = Sub.to_list sub in
  let append s (x,t) =
   T.to_stringm t >>= fun t ->
   find_var_name x >>= fun x ->
   return (s^" "^x^"->"^t)
  in foldl append "" sub
 ;;

let print_subs ss =
 let add s sub = print_sub sub >>= fun s' -> return (s^s'^"\n ") in
 foldl add " " ss >>= fun s ->
 Format.printf "%s%!" s; return ()
;;

 (* flatten term and sort AC arguments (order is crucial for
    algorithm by Christian/Lincoln) *)

   (* sorts list of terms such that constants come before compound
    terms where the root has positive arity come before variables *)
 let rec lex = function
    [],[] -> 0
  | [],_ -> -1
  |_, [] -> 1
  | x::xs,y::ys -> let c = ucmp x y in if c=0 then lex (xs,ys) else c
 and ucmp t t' =
  match t, t' with
     T.Var x, T.Var y -> Var.compare x y
   | T.Fun _, T.Var _ -> -1
   | T.Var _, T.Fun _ -> 1
   | T.Fun(_, []), T.Fun(_, t::ts) -> -1
   | T.Fun(_, t::ts), T.Fun(_, []) -> 1
   | T.Fun(f, fs), T.Fun(g, gs) when f = g -> lex (fs, gs)
   | T.Fun(f, _), T.Fun(g, _) -> Fun.compare f g
 ;;

 let sort_flatten t =
  T.flatten t >>= function
   | T.Fun(f, ts) -> return (T.Fun(f, List.sort ucmp ts))
   | v -> return v
	;;

 (* removes common occurrences in both argument lists (like mutual
    multiset difference *)
 let remove_common_args xs ys =
  let rec remove (sx,sy) = function
   | [], ys -> List.rev sx, List.rev sy @ ys
   | xs, [] -> List.rev sx @ xs, List.rev sy
   | x::xs,y::ys when x=y -> remove (sx,sy) (xs,ys)
   | x::xs,y::ys when ucmp x y < 0 -> remove (x::sx,sy) (xs,y::ys)
   | x::xs,y::ys -> remove (sx,y::sy) (x::xs,ys)
  in remove ([],[]) (xs,ys)
 ;;

 (* * *          auxiliary functions for Fortenbacher's algorithm         * * *)
 (* For s=f(s_1,...,s_m) and t=f(t_1,...,t_n) for AC-Symbol f, abstract s t
    returns a pair (f(x_1,...,x_m), f(y_1,...,y_n)) together with pairs
    (x_i,s_i) and (y_j,t_j) where x_i, y_j are fresh variables.*)
 let abstract all_fresh f ss ts =
  let var t =
   match t with
    T.Var x when all_fresh -> return x
   | u -> fresh_var >>= fun x' -> return x'
  in
  map var ss >>= fun xs -> let xst = List.map T.make_var xs in
  map var ts >>= fun ys -> let yst = List.map T.make_var ys in
  let c0 = (T.Fun(f,xst), T.Fun(f,yst)) in
  return (c0, (List.zip xst ss) @ (List.zip yst ts),xs,ys)
 ;;
 
 (* fill matrix with T.Var vi lists according to solutions *)
 let to_semigroup f ass =
  let ith v i = List.replicate i (T.Var v) in
  let sg a (acc, vs) =
   fresh_var >>= fun vi ->
   let a' = List.map (fun (x,i) -> ith vi i) a in
   return (a'::acc,vi::vs)
  in
  foldr sg ([],[]) ass >>= fun (ass,vs) ->
  return (A.of_list (List.map A.of_list ass), vs)
 ;;

 (* Construct all subsets of the assignment set b such that there is no zero
    column sum. Some constraints:
    - for a column corresponding to a non-variable term, the column sum is 1
    - delete rows where ones are at positions corresponding to non-unifiable
      terms (only implemented partially here)
    where the two conditions are checked by check_row *)
 let check_row terms b i =
  let a = List.nth b i in
  assert(List.length a = (List.length terms));
  let ts = List.zip_with (fun t (_,value) -> t,value) terms a in
  let compound_true (t,v) = (v > 0) && (not (T.is_var t)) in
  let ts = List.filter compound_true ts in
  (* no entries > 1 *)
  if List.for_all (fun (_,v) -> v=1) ts then
   (* no nonunifiable entries *)
   List.for_all (fun ((t,_),(u,_)) -> T.root t = (T.root u)) (List.square ts)
  else false
 ;;

 let valid m terms subset =
  let col i =
   let s = List.fold_left (fun s j -> s + m.(j).(i)) 0 subset in
   if not (T.is_var (List.nth terms i)) then s = 1 else s > 0
  in List.for_alli (fun i _ -> col i) terms
 ;;

 let subsets (ss,ts) b =
  (* filter out equal variable *)
  let cmp t s = if T.is_var t && (T.is_var s) then T.compare t s else 1 in
  let terms = List.unique ~c:cmp (ss@ts) in (* same order as variables *)
  let indices = List.range 0 (List.length b) in
  let indices = List.filter (check_row terms b) indices in
  (* matrix now more convenient *)
  let b' = List.map (List.map snd) b in
  let m = A.of_list (List.map A.of_list b') in
  (* possible index subsets *)
  let ixsets = List.remove [] (List.powerset indices) in
  (* check *)
  List.filter (valid m terms) ixsets
 ;;

 let sum_up subsets b' vars f =
  let vars = List.unique vars in
  let fterm f = function [x] -> x | args -> T.Fun(f,args) in
  let col_sum sub i = fterm f (List.flat_map (fun r -> b'.(r).(i)) sub) in
  let add_binding sub i s v = Sub.add v (col_sum sub i) s in
  let sub_for_subset s = List.foldli (add_binding s) Sub.empty vars in
  List.map sub_for_subset subsets
 ;;

 (* make substitution idempotent *)
 let make_idempotent sigma =
  let rec fix sigma sigman =
   let sigman' = Sub.compose Sub.apply_term sigman sigma in
   if Sub.equal sigman' sigman then sigman
   else fix sigma sigman'
  in fix sigma sigma
 ;;

 (* * *                      main unification procedure                   * * *)
 (* Given terms s and t, return complete set of unifiers for s=_{AC} t.
    Calls unify' and simplifies returned set of bindings in that all
    bindings x -> t are removed where x does not occur in s or t. *)
 let rec unify s t =
  (* restrict substitution to variables of interest *)
  let vars = List.union (T.vars s) (T.vars t) in
  let add x t s = if List.mem x vars then Sub.add x t s else s in
  let simplify s = Sub.fold add s Sub.empty in
  (liftM (Option.map (List.map simplify))) (unify' s t)
 (* Given terms s and t, return complete set of unifiers for s=_{AC} t.
    When matching, s is ground *)
 and unify' s t =
  match s, t with
   | T.Var x, _ ->
    if T.is_proper_subterm s t then return None (* occur check *)
    else if s = t then return (Some zero_list)
    else return (Some [Sub.add x t Sub.empty])
   | _, T.Var y ->
    if T.is_proper_subterm t s then return None (* occur check *)
    else return (Some [Sub.add y s Sub.empty])
   | T.Fun(a,[]),T.Fun(b,[]) -> (* both constants *)
    if a = b then return (Some zero_list) else return None
   | T.Fun(f, _), T.Fun(g, _) when f <> g -> return None
   | T.Fun(f,ss), T.Fun(_, ts) -> ((* same root symbol *)
    is_ac_symbol f >>= fun f_is_ac ->
    if not f_is_ac then (* assume length ss = length ts *)
     unify_with_list (List.zip ss ts) zero_list
    else
     sort_flatten s >>= fun s' -> sort_flatten t >>= fun t' ->
     let ss', ts' = remove_common_args (T.args s') (T.args t') in
     if List.is_empty ss' && List.is_empty ts' then
		  return (Some [Sub.empty])
     else if List.is_empty ss' || List.is_empty ts' then
      return None
     else
      (* (1) abstract non-variable terms *)
      abstract true f ss' ts' >>= fun (_,cs,xs,ys) ->
      (* (2) solve diophantine equations *)
      let b = Dio.solve xs ys in
      (* (3) make semigroup representation, fresh vars for rows *)
      to_semigroup f b >>= fun (b',vars) ->
      (* (4) find usable subsets of rows in b *)
      let bsubsets = subsets (ss',ts') b in
      (* (5) construct semigroup unifiers *)
      let gamma = sum_up bsubsets b' (xs @ ys) f in
      (* (6) make substitutions idempotent *)
      let theta = List.map make_idempotent gamma in
      unify_with_list cs theta
     )
 (* Given list (i.e. conjunction) of equations es+ set of substitutions
    subs, take union of CSU(es\theta) for all \theta in subs,
    where CSU(X) is complete set of unifiers for X. *)
 and unify_with_list es subs =
  let add csus theta =
   let compose s = Sub.compose Sub.apply_term theta s in
   let cs = List.map (Pair.map (Sub.apply_term theta)) es in
   unify_conjunction cs >>= function
   | None -> return csus
   | Some z -> return (List.rev_append (List.map compose z)  csus)
  in foldl add [] subs >>= function
     [] -> return None
	 | subs -> return (Some subs)

 (* Given a list (i.e. conjunction) of equations es, return complete set
    of unifiers for es. *)
 and unify_conjunction cj =
  match cj with
  | [] -> return (Some zero_list) (*e.g. unify (fxy,fxy) *)
  | [s,t] -> unify s t
  | (s, t) :: es -> unify s t >>= function
   | None -> return None | Some ss -> unify_with_list es ss
 ;;

 let unify s t =
   unify s t >>= function
      Some ss -> return (List.unique ~c:Sub.compare ss)
    | _ -> return []
 ;;

 let are_unifiable s t =
	unify s t >>= function [] -> return false | _ -> return true
 ;;

 let rec tsub_apply tsub = function
  | T.Fun(c,[]) as ct -> TSub.apply ct ct tsub
  | T.Fun(f,ts) -> T.Fun(f,List.map (tsub_apply tsub) ts)
  | v -> v
 ;;

 let reverse_skolemization rho =
  let tsub_add x t s = TSub.add t (T.Var x) s in
  Sub.fold tsub_add rho TSub.empty
 ;;

 let skolemization =
  let rec skolemization s = function
   | T.Var x ->
    if Sub.mem x s then return s
    else M.fresh_fun >>=
     M.create_fun_name >>= M.create_fun 0 >>= (fun c ->
     return (Sub.add x (T.Fun(c,[])) s))
   | T.Fun (f,ts) -> M.foldl skolemization s ts
  in skolemization Sub.empty
 ;;

 (* Returns a complete set of matching substitutions \sigma such that for
    every \theta in \sigma t\theta=s. *)
 let match_term s t =
  skolemization s >>= fun rho ->
  let s' = Sub.apply_term rho s in
  unify s' t >>= fun sigmas ->
  let rho_inv = reverse_skolemization rho in
  return (List.map (Sub.map (tsub_apply rho_inv)) sigmas)
 ;;

 let matches s t =
	match_term s t >>= fun ss ->
	return (ss <> [])
 ;;

 (* TESTS *)
 let test () =
  let sigma = Sig.empty 20 in
  let x,sigma = Sig.create_var "x" sigma in
  let y,sigma = Sig.create_var "y" sigma in
  let z,sigma = Sig.create_var "z" sigma in
  let u,sigma = Sig.create_var "u" sigma in
  let v,sigma = Sig.create_var "v" sigma in
  let w,sigma = Sig.create_var "w" sigma in
  let x,y,z,u,v,w = T.Var x, T.Var y, T.Var z, T.Var u, T.Var v, T.Var w in
  let f,sigma = Sig.create_fun 2 "f" sigma in
  let sigma = Sig.add_theory f Theory.AC sigma in
  let h,sigma = Sig.create_fun 2 "h" sigma in
  let sigma = Sig.add_theory h Theory.AC sigma in
  let a,sigma = Sig.create_fun 0 "a" sigma in
  let b,sigma = Sig.create_fun 0 "b" sigma in
  let c,sigma = Sig.create_fun 0 "c" sigma in
  let d,sigma = Sig.create_fun 0 "d" sigma in
  let e,sigma = Sig.create_fun 0 "e" sigma in
  let g,sigma = Sig.create_fun 1 "g" sigma in
  let a_ = T.Fun(a, []) in
  let b_ = T.Fun(b, []) in
  let c_ = T.Fun(c, []) in
  let d_ = T.Fun(d, []) in
  let e_ = T.Fun(e, []) in
  let faa = T.Fun(f, [a_;a_]) in
  let faaa = T.Fun(f, [a_;faa]) in
  let fabc = T.Fun(f, [a_;T.Fun(f, [b_;c_])]) in
  let fcab = T.Fun(f, [c_;T.Fun(f, [a_;b_])]) in
  let faby = T.Fun(f, [a_;T.Fun(f, [b_;y])]) in
  let fxy = T.Fun(f, [x;y]) in
  let fxx = T.Fun(f, [x;x]) in
  let fyy = T.Fun(f, [y;y]) in
  let fzz = T.Fun(f, [z;z]) in
  let fax = T.Fun(f, [a_;x]) in
  let gx = T.Fun(g, [x]) in
  let ga = T.Fun(g, [a_]) in
  let gb = T.Fun(g, [b_]) in
  let gu = T.Fun(g, [u]) in
  let fgax = T.Fun(f, [ga; x]) in
  let fgay = T.Fun(f, [ga; y]) in
  let gfaa = T.Fun(g,[faa]) in
  let gfax = T.Fun(g,[fax]) in
  let t2 = T.Fun(f, [T.Fun(h,[z;v]);T.Fun(h,[u;v]);w]) in
  let rec flist = function
    [s; t] -> T.Fun(f, [s; t])
   | s :: ts -> T.Fun(f, [s; flist ts])
   | _ -> failwith "Unexpected pattern"
  in
  let t_lc88_1 = flist [c_;c_;gu; x] in
  let t_lc88_2 = flist [b_;ga;y; z] in
  (* testing unification *)
  let run_unify (s,t) = Either.right (M.run sigma (
   T.flatten s >>= fun s ->
   T.flatten t >>= fun t ->
   unify s t )) in
  let print_subs (s,t) = let (s,n) = Either.right (M.run sigma (
   T.flatten s >>= fun s ->
   T.flatten t >>= fun t ->
   T.to_stringm s >>= fun ss ->
   T.to_stringm t >>= fun ts ->
   let m = "To unify "^ss^" and "^ts^", " in
   unify s t >>= function
     [] -> return (m^"none", 0)
   | us ->
    let m' = m^(string_of_int (List.length us))^" substitutions\n" in
    foldl
     (fun s u -> print_sub u >>= fun us -> return (s^"\n or\n"^us))
     m' us >>= fun s -> return (s, List.length us)
   )) in Format.printf "%s\n" s; n in
  let assert_some ts =
   assert(ignore (print_subs ts); run_unify ts <> []) in
  let assert_more ts n =
   assert(run_unify ts <> []); assert (print_subs ts = n) in
  let assert_none ts = assert(run_unify ts == []) in
  assert_some (x,x);
  assert_some (fxy,t2);
  assert_some (x,y);
  assert_some (x,a_);
  assert_some (a_, y);
  assert_some (a_,a_);
  assert_none (ga,gb);
  assert_some (gx,ga);
  assert_some (faa, y);
  assert_some (gx,gfaa);
  assert_none (a_, gx);
  assert_none (x, gx);
  assert_some (y, gx);
  assert_none (faa, faaa);
  assert_some (fcab, fabc);
  assert_some (faa, fax);
  assert_some (gfaa, gfax);
  assert_some (faaa, fax);
  assert_some (fabc, fax);
  assert_some (faaa, fxy);
  assert_some (fabc, fxy);
  assert_some (faby, fxy);
  assert_some (T.Fun(f,[u; gfaa]), T.Fun(f, [gfax; y]));
  assert_none (fgax, fax);
  assert_some (fgay, fax);
  assert_none (fgay, faaa);
  assert_some (fgay, fgay);
  assert_some (T.Fun(f, [T.Fun(g,[faa]); u]), faby);
  (* some examples from christian/lincoln 88 *)
  assert_some (t_lc88_1,t_lc88_2);
  assert_some (flist [w;x;gx],flist [y;u;gu]);
  assert_more (flist [x;a_;b_],flist [u;c_;d_;e_]) 2;
  assert_more (flist [x;a_;b_],flist [u;c_;c_;d_]) 2;
  assert_more (flist [x;a_;b_],flist [u;c_;c_;c_]) 2;
  assert_more (flist [x;a_;b_],flist [u;y;c_;d_]) 12;
  assert_more (flist [x;a_;b_],flist [u;y;c_;c_]) 12;
  assert_more (flist [x;a_;b_],flist [u;y;z;c_]) 30;
  assert_more (flist [x;a_;b_],flist [u;y;z;v]) 56;
  assert_more (flist [x;a_;a_],flist [u;c_;d_;e_]) 2;
  assert_more (flist [x;a_;a_],flist [u;c_;c_;d_]) 2;
  assert_more (flist [x;a_;a_],flist [u;c_;c_;c_]) 2;
  assert_more (flist [x;a_;a_],flist [u;y;c_;d_]) 8;
  assert_more (flist [x;a_;a_],flist [u;y;v;c_]) 18;
  assert_more (flist [x;a_;a_],flist [u;y;z;v]) 32;
	assert_more (flist [x;y;a_],flist [u;c_;d_;e_]) 28;
	assert_more (flist [x;y;a_],flist [u;c_;c_;d_]) 20;
	assert_more (flist [x;y;a_],flist [u;v;c_;d_]) 88;
	assert_more (flist [x;y;a_],flist [u;v;z;c_]) 204;
	assert_more (flist [x;y;z],flist [u;v;c_;d_]) 336;
	assert_more (flist [x;y;z],flist [u;v;w;c_]) 870;
	(* examples with repeated variables *)
	assert_some (T.Fun(f,[x]), T.Fun(f,[y]));
	assert_some (fxy,fzz);
	assert_some (fxx,fyy);
	assert_some (flist [x;x;gx], flist[u;u;gu]);
 (* testing matching *)
 let run_matches (s,t) = Either.right (M.run sigma (match_term s t)) in
 let print_msubs (s,t) = let (s,n) = Either.right (M.run sigma (
  T.to_stringm s >>= fun ss ->
  T.to_stringm t >>= fun ts ->
  let m = "To match "^ss^" with "^ts^", " in
  match_term s t >>= function
    [] -> return ("none", 0)
  | us ->
   let m' = m^(string_of_int (List.length us))^" substitutions\n" in
   foldl
    (fun s u -> Sub.to_stringm u >>= fun us -> return (s^"\n or\n"^us))
    m' us >>= fun s -> return (s, List.length us)
  )) in Format.printf "%s\n" s; n in
  let assert_msome ts =
   assert([] <> run_matches ts); ignore (print_msubs ts) in
  let assert_mnone ts = assert([] == (run_matches ts)) in
  assert_msome (x,x);
	assert_mnone (x,a_);
	assert_msome (a_, y);
	assert_msome (a_,a_);
	assert_mnone (gx,ga);
	assert_msome (faa, y);
	assert_msome (gfaa,gx);
	assert_mnone (a_, gx);
	assert_msome (gx, x); (* ! - but with renamed rules no problem *)
	assert_mnone (y, gx);
	assert_mnone (faa, faaa);
	assert_msome (fcab, fabc);
	assert_msome (faa, fax);
	assert_msome (gfaa, gfax);
	assert_msome (faaa, fax);
	assert_msome (fabc, fax);
	assert_msome (faaa, fxy);
	assert_msome (fabc, fxy);
	assert_msome (faby, fxy);
	assert_msome (T.Fun(f,[u; gfaa]), T.Fun(f, [gfax; y]));
	assert_msome (flist[ga;u;w], fxy);
	assert_mnone (fgay, fax);
	assert_mnone (fgay, faaa);
	assert_msome (fgay, fgay);
	assert_msome (flist [a_;ga;b_;x], flist[a_;y;ga;b_])
 ;;
end
