(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** MODULES ******************************************************************)
module M = Monad;;

(*** OPENS ********************************************************************)
open Prelude;;
open Util;;

(*** MODULES ******************************************************************)
module Make (L : LABEL) = struct
 (*** MODULES *****************************************************************)
 module P = Position;;
 module Var = Variable;;
 module Sig = Signature.Make (L);;
 module T = Term.Make (L);;
 module C = Context.Make (L);;
 module E = Elogic.Make (L);;
 module Monad = M.Make (L);;
 module R = Rule.Make (L);;
 module Sub = Substitution.Make (L);;
 module Trs = Trs.Make (L);;
 module L = Aclogic.Make (L);;

 (*** TYPES *******************************************************************)
 type 'a m = 'a Monad.t;;
 type term = T.t;;
 type rule = R.t;;
 type trs = Trs.t;;
 type position = P.t;;

 type overlap = {
	 inner: R.t;
   pos: position;
	 outer: R.t;
	 sub: Sub.t
 }

 type context = {
  match_cache         : (T.t * T.t, Sub.t list) Hashtbl.t;
  reducts_cache       : (T.t, T.t list) Hashtbl.t;
  state               : Sig.t;
  unify_cache         : (T.t * T.t, Sub.t list) Hashtbl.t;
 }

 (*** MODULES (part 2) ********************************************************)
 module Statex = struct type t = context end;;
 module M = Util.Monad.Transformer.State (Statex) (Monad);;
 
 (*** FUNCTIONS ***************************************************************)
 let mk_context sigma = {
  match_cache = Hashtbl.create 256;
  reducts_cache = Hashtbl.create 256;
  state = sigma;
  unify_cache = Hashtbl.create 256;
 }

 let return = M.return;;
 let (>>=) = M.(>>=);;
 let (>=>) = M.(>=>);;
 let liftM = M.liftM;;
 let lift = M.lift;;

 let cache tbl f k = 
  if Hashtbl.mem tbl k then return (Hashtbl.find tbl k)
  else (f k >>= fun v -> (Hashtbl.add tbl k v; return v))
 ;;

 let cache2symm tbl f k1 k2 = 
  let k, k' = (k1, k2), (k2, k1) in
  if Hashtbl.mem tbl k then return (Hashtbl.find tbl k)
  else if Hashtbl.mem tbl k' then return (Hashtbl.find tbl k)
  else (uncurry f k >>= fun v -> (
    Hashtbl.add tbl k v; Hashtbl.add tbl k' v; return v))
 ;;

 let cache2l tbl f k1 k2 = cache tbl (lift @.@ (uncurry f)) (k1, k2);;

 (*constructors for step*)
 let overlap_make i p o s = {
   inner = i; 
   pos = p; 
   outer = o;
   sub = s
 };;

 (* lift into M *)
 let is_ac = lift @.@ (Monad.is_theory Theory.AC);;
 let fresh_var = lift Monad.fresh_var;;
 let flatten = lift @.@ T.flatten;;

 (* caching unification/matching speeds up by > 40% on wcr for groups/rings *)
 let match_term t l = M.get >>= fun c-> cache2l c.match_cache L.match_term t l;;
 let unify s t =
   M.get >>= fun c ->
   cache2symm c.unify_cache (fun s -> lift @.@ (L.unify s)) s t
 ;;

 (* return all pairs (ys,zs) such that ys \cup zs = xs and both ys and zs are
    nonempty. (Symmetries not eliminated, i.e., both (ys,zs) and (zs,ys) occur
		in the result) *)
 let bipartition xs =
  let rec bipartition ys = function
   | [] -> ys
   | x::xs -> bipartition (List.flat_map (fun (a,b) -> [x::a,b; a,x::b]) ys) xs
  in List.filter (fun (ys,zs) -> ys<>[] && zs<>[]) (bipartition [[],[]] xs)
 ;;

 (* rename variables in term, extending substitution sub and with next variable
    v. Returns renamed term, extended substitution, and the next free variable*)
 let rec normalize sub v = function
  | T.Var x -> (
    try (Sub.find x sub, sub, v)
    with Not_found -> (T.Var v, Sub.add x (T.Var v) sub, Var.next v))
  | T.Fun (f, ts) ->
   let fold_norm (l,s,v) t = let (t',s,v) = normalize s v t in (t'::l,s,v) in
   let (ts', sub', v') = List.fold_left fold_norm ([], sub, v) ts in
   (T.Fun (f, List.rev ts'), sub', v')
 ;;

 (* rename variables in term pair to improve caching (and readability) *)
 let normalize_terms (t, u) =
  let t', s, v = normalize Sub.empty Var.zero t in
  let u', _, _ = normalize s v u in
  (t', u')
 ;;

 (* flatten term and sort arguments of AC symbols in some fixed way *)
 let sort_flatten t = 
  let rec sort = function
   | T.Fun(f,ts) ->
    M.map sort ts >>= fun ts' ->
    lift (Monad.is_theory Theory.AC f) >>= fun f_is_ac ->
    M.return (T.Fun (f, if f_is_ac then List.sort compare ts' else ts'))
   | v -> M.return v
  in flatten t >>= sort
 ;;

 (* make term of AC symbol f and (arbitrary but positive number of) arguments *)
 let rec fterm f = function 
	   [t] -> t 
	 | [_; _] as ts -> T.Fun(f, ts)
	 | t::ts -> T.Fun(f, [t; fterm f ts])
	 | _ -> failwith "Acrewrite.fterm: empty argument list not allowed"
  ;;

 (* rearrange arguments of AC symbols for canonical representation *)
 let rec ac_normalize t = sort_flatten t >>= function 
  | T.Fun(f,ts) as t -> 
	 is_ac f >>= fun f_is_ac ->
	 if not f_is_ac then return t
   else M.map ac_normalize ts >>= fun ts' -> return (fterm f ts')
  | v -> return v
 ;;

 (* check whether two terms are AC equivalent *)
 let acequivalent t u =
   M.project sort_flatten (t,u) >>= (return @.@ (Util.uncurry T.equal))
 ;;

 (* enumerate all terms in AC-equivalence class with root f and given argument
    list. Final unique_hash makes a difference only if ts has duplicates  *)
 let rec acequivalents f ts = 
	let symm ts = [ T.Fun(f, ts); T.Fun(f, List.rev ts) ] in 
	let rec aceqs = function
   | []
   | [_] as ts -> ts
   | [_; _] as ts -> symm ts
   | ts ->
    let combine (t,u) =
 	   let ts, us = Pair.map aceqs (t,u) in
	   List.flat_map (fun (t,u) -> symm [t;u]) (List.product ts us)
    in List.flat_map combine (bipartition ts)
	 in List.unique_hash (aceqs ts)
 ;;

 (* given term t, return list pts of pairs (p, u) such that u is AC equivalent
    to t and p is a position in u. It is complete in the sense that whenever
		p and t' satisfy this condition then there is some (p', u') in pts such
		that u|_p is the same as u'|_p' and p and p' coincide on the non-AC part. *)
 let rec ac_pos p = function
  | T.Var _ as x -> return (if p x then [P.root, x] else [])
  | T.Fun (f,ts) as t ->
	 is_ac f >>= fun f_is_ac ->
	 if not f_is_ac then
	  (* just add argument numbers to positions in subterms *)
	  let add i (p,u) = P.add_first i p,C.apply u (C.of_term (P.of_list [i]) t) in
	  M.flat_mapi (fun i -> liftM (List.map (add i)) @.@ ac_pos p) ts >>= fun ps->
		return (if p t then (P.root, t)::ps else ps)
	 else
	  flatten t >>= (return @.@ T.args) >>= fun tfs ->
		let add_partition (xs,zs) = match xs with
		 (* for singleton partition [ti] add all positions in ti *)
		  [ti] -> ac_pos p ti >>= M.map (fun (p,u) -> return (0::p,fterm f (u::zs)))
		 (* non-singleton partitions give rise to AC term *)
		 | _ -> let ti = T.make_fun f [fterm f xs;fterm f zs] in
            return (if p ti then [P.of_list [0], ti] else [])
		in
		(* split argument set of flattened term into two *)
	  M.flat_map add_partition (bipartition tfs) >>=
		(return @.@ List.unique_hash @.@ (if p t then List.cons (P.root,t) else id))
 ;;

 let ac_funs_pos = ac_pos (T.is_fun);;
 let ac_vars_pos = ac_pos (T.is_var);;
 let ac_pos = ac_pos (fun _ -> true);;

 (* positions relevant for overlaps with extension rules: only those covering
    the extension variable (supposed to be first argument) *)
 let xpos = function
  | T.Fun (f, [x; t2]) as t ->
	 is_ac f >>= fun f_is_ac ->
	 if (not f_is_ac) || not (T.is_var x) then
    failwith "Acrewrite.xpos requires argument to be result of AC extension"
   else
	  flatten t2 >>= (return @.@ T.args) >>= fun t2s ->
    let fterm xs zs = T.make_fun f [fterm f (x :: xs); fterm f zs] in
		let for_partition (xs,zs) = return (P.of_list [0], fterm xs zs) in
	  M.map for_partition (bipartition t2s) >>=
		(return @.@ List.unique_hash @.@ (List.cons (P.root,t)))
  | _ -> failwith "Acrewrite.xpos: invalid argument"
 ;;

 (* AC rewrite term t at position p with rule; returns list of reducts, which is
    empty if no step is possible. The result is AC-sorted, which is crucial
    for e.g. reduct computation to avoid blowup  *)
 let rewrite_at t rule p =
	let c = C.of_term p t in
  let s = T.subterm p t in
  let l, r = R.to_terms rule in
  let rewrite_with sigma = return (C.apply (Sub.apply_term sigma r) c) in
  match_term s l >>= M.map (rewrite_with >=> ac_normalize)
 ;;

 (* single-step AC-reducts from t using rules in trs. Caching results yields
    ~ 30% speedups when checking wcr on AC rings and groups *)
 let direct_reducts t trs =
  M.get >>= fun c ->
  let compute t =
	 ac_funs_pos t >>= fun ps ->
	 let with_rl rl = M.flat_map (fun (p, t) -> rewrite_at t rl p) ps in
	 M.flat_map with_rl (Trs.to_list trs) >>=
	 (return @.@ List.unique_hash)
  in cache c.reducts_cache compute t
 ;;

 (* auxiliary for reducts and joinable: rewrites terms in ts one step,
    returning pair (ts', acc') where ts' contains newly derived terms and
    acc' extends acc by ts *)
 let direct_reducts_list trs width (ts,acc) =
  M.flat_map (flip direct_reducts trs) ts >>= fun rs ->
  let rs = if width < 0 then rs else List.take width (List.unique_hash rs) in
  let acc' = acc @ ts in
  let ts' = List.diff rs acc' in
 return (ts', acc')
 ;;

 let rec reducts' n width acc ts trs = match n with
  | 0 -> return acc
  | n -> (
   direct_reducts_list trs width (ts, acc) >>= fun (ts', acc') ->
   if ts' = [] then return acc'
   else reducts' (n-1) width (acc'@ts') ts' trs
  )
 ;;

 let reducts ?(n = ~-1) ?(width = ~-1) t trs = reducts' n width [] [t] trs

 let rec are_joinable' n width (us, uacc) (ts, tacc) trs = 
  let joined ts = M.exists (fun u -> M.exists (acequivalent u) ts) in
  let step = direct_reducts_list trs width in
  match n with
   | 0 -> joined (ts@tacc) (us@uacc)
   | n ->
  M.project step ((ts, tacc), (us, uacc)) >>= fun ((ts',tacc'), (us',uacc'))  ->
  joined (ts'@tacc') (us'@uacc') >>= fun b ->
  if b || (ts' = [] && us' = []) then return b
  else are_joinable' (n-1) width (ts',tacc') (us',uacc') trs
 ;;

 let are_joinable ?(n = ~-1) ?(width = ~-1) u t trs =
  are_joinable' n width ([u],[]) ([t],[]) trs
 ;;

 let is_normalform t trs = direct_reducts t trs >>= (return @.@ ((==) []))

 (* find an AC normal form in a given list of terms *)
 let find_normalform ?(n = ~-1) term trs =
  reducts ~n:n term trs >>= fun rds ->
  let find r t = 
   match r with
    | Some _ as nf -> return nf
    | _ -> is_normalform t trs >>= fun b ->
     if b then return (Some t) else return None
  in M.foldl find None (term :: rds) >>= function
   | None -> return (match rds with r :: _ -> r | _ -> term)
   | Some u -> return u
 ;;

 (* check whether <inner, p, outer> allows for an AC overlap, where p is a
    position in the lhs of outer *)
 let check_overlap inner p outer =
  let s = T.subterm p (R.lhs outer) in
  unify s (R.lhs inner) >>=
  M.map (return @.@ (overlap_make inner p outer))
 ;;

let has_linear_var_as_flat_arg rl =
 M.project flatten (R.to_terms rl) >>= fun (l,r) ->
 let f = Option.the (T.root l) in
 let ls,rs = T.args l, T.args r in
 let is_lin_var ti ts =
  let tis, ts' = List.partition ((=) ti) ts in
  let as_sub = List.exists (fun tj -> List.mem ti (T.subterms tj)) ts' in
  T.is_var ti && List.length tis = 1 && not as_sub
 in
 let is_lin_var_rhs ti = r=ti || T.root r = Some f && is_lin_var ti rs in
 return (List.exists (fun x -> is_lin_var x ls && is_lin_var_rhs x) ls)
;;

(* return Some extension (if its lhs is rooted by an AC symbol, and no linear
   variable is present) or otherwise None *)
let extend rl = 
 let f = Option.the (T.root (R.lhs rl)) in
 is_ac f >>= fun f_is_ac ->
 has_linear_var_as_flat_arg rl >>= fun has_lin_var ->
 if not f_is_ac || has_lin_var then return None
 else
  fresh_var >>= (return @.@ T.make_var) >>= fun y ->
  let rl = R.of_terms (T.Fun(f,[y; R.lhs rl])) (T.Fun(f,[y; R.rhs rl])) in
	return (Some rl)
 ;;

 (* compute AC overlaps among (extensions of) rules in trs *)
 let overlaps trs =
	M.map (lift @.@ R.rename) (Trs.to_list trs) >>= fun ro ->
  let fpos r = ac_funs_pos (R.lhs r) in
  M.map (fun r -> fpos r >>= fun p -> return (r,p)) ro >>= fun rop ->
  (* for extended rules positions involving fresh variable suffice *)
	M.flat_map (liftM (function Some r -> [r] |_ -> [])@.@extend) ro >>= fun rx ->
  M.map (fun r -> xpos (R.lhs r) >>= fun ps -> return (r, ps)) rx >>= fun rxp ->
	(* combinations of outer and inner rule: extensions only required for outer *)
	let rpairs = List.product (rxp @ rop) (Trs.to_list trs) in
	let of_rls ((o, ps),i) =
		(* l' is AC equivalent to original lhs of outer *)
		let mk_overlap (p,l') = check_overlap i p (R.of_terms l' (R.rhs o)) in
		M.flat_map mk_overlap ps
	in
	M.flat_map of_rls rpairs >>=
  (return @.@ List.unique_hash)
 ;;

 (* make AC critical peak from AC overlap *)
 let cpeak_from_overlap o =
   let c = C.of_term o.pos (R.lhs o.outer) in
   let sub = Sub.apply_term o.sub in
   let u = C.apply (sub (R.rhs o.inner)) (Sub.apply_context o.sub c) in
   let w = sub (R.rhs o.outer) in
   let v = sub (R.lhs o.outer) in
   (u,v,w)
 ;;

 (* make AC critical pair from AC overlap *)
 let cp_from_overlap o =
   let (u,v,w) = cpeak_from_overlap o in normalize_terms (u,w)
 ;;

 let cps trs = overlaps trs >>= M.map (return @.@ cp_from_overlap)

 let cpeaks trs = overlaps trs >>= M.map (return @.@ cpeak_from_overlap)

 let is_wcr ?(n = ~-1) trs =
  cps trs >>= fun cps ->
  let cps = List.map (fun (s,t) -> if s < t then (s,t) else (t,s)) cps in
  M.for_all (flip (uncurry (are_joinable ~n:n ~width:8)) trs) (List.unique_hash cps)
 ;;

 (* run on plain signature monad *)
 let (>>=) = Monad.(>>=);;
 let return = Monad.return;;
 let run m a state = M.run (mk_context state) (m a);;
 let ac_pos t = Monad.get >>= run ac_pos t;;
 let direct_reducts t trs = Monad.get >>= run (direct_reducts t) trs;;
 let acequivalent s t = Monad.get >>= run (acequivalent s) t;;
 let cps trs = Monad.get >>= run cps trs;;
 let cpeaks trs = Monad.get >>= run cpeaks trs;;
 let is_normalform t trs = Monad.get >>= run (is_normalform t) trs;;
 let ac_normalize t = Monad.get >>= run ac_normalize t;;

 let find_normalform ?(n = ~-1) t trs =
  Monad.get >>= run (find_normalform ~n:n t) trs
 ;;

 let are_joinable ?(n = ~-1) ?(width = ~-1) s t trs =
  Monad.get >>= run (are_joinable ~n:n ~width:width s t) trs
 ;;

 let reducts ?(n = ~-1) ?(width = ~-1) t trs =
  Monad.get >>= run (reducts ~n:n ~width:width t) trs
 ;;

 let is_wcr ?(n = ~-1) trs = Monad.get >>= run (is_wcr ~n:n) trs;;

 (* Checks if term u reaches in n AC rewrite steps of TRS trs the term v *)
 let is_reachable ?(n = ~-1) u v trs = 
  reducts ~n:n u trs >>= fun rds -> return (List.mem v rds)
 ;;

 (* TESTS *)
 let test () = 
  let sigma = Sig.empty 20 in
  let x,sigma = Sig.create_var "x" sigma in
  let y,sigma = Sig.create_var "y" sigma in
  let z,sigma = Sig.create_var "z" sigma in
  let u,sigma = Sig.create_var "u" sigma in
  let v,sigma = Sig.create_var "v" sigma in
  let w,sigma = Sig.create_var "w" sigma in
  let x,y,z,u,v,w = T.Var x, T.Var y, T.Var z, T.Var u, T.Var v, T.Var w in
  let f,sigma = Sig.create_fun 2 "f" sigma in
  let sigma = Sig.add_theory f Theory.AC sigma in
  let a,sigma = Sig.create_fun 0 "a" sigma in
  let b,sigma = Sig.create_fun 0 "b" sigma in
  let c,sigma = Sig.create_fun 0 "c" sigma in
  let d,sigma = Sig.create_fun 0 "d" sigma in
  let e,sigma = Sig.create_fun 0 "e" sigma in
  let g,sigma = Sig.create_fun 1 "g" sigma in
  let h,sigma = Sig.create_fun 2 "h" sigma in
  let zero,sigma = Sig.create_fun 0 "0" sigma in
  let one,sigma = Sig.create_fun 0 "1" sigma in
  let m,sigma = Sig.create_fun 2 "m" sigma in
  let sigma = Sig.add_theory m Theory.AC sigma in
  let a_ = T.Fun(a, []) in
  let b_ = T.Fun(b, []) in
  let c_ = T.Fun(c, []) in
  let d_ = T.Fun(d, []) in
  let e_ = T.Fun(e, []) in
  let zero = T.Fun(zero, []) in
  let one = T.Fun(one, []) in
  let faa = T.Fun(f, [a_;a_]) in
  let fab = T.Fun(f, [a_;b_]) in
  let fac = T.Fun(f, [a_;c_]) in
  let fxx = T.Fun(f, [x;x]) in
  let ga = T.Fun(g, [a_]) in
  let gu = T.Fun(g, [u]) in
  let gx = T.Fun(g, [x]) in
  let faaa = T.Fun(f, [a_;faa]) in
  let gfaa = T.Fun(g, [faa]) in
  let faaaa = T.Fun(f, [faa;faa]) in
  let fabc = T.Fun(f, [a_;T.Fun(f, [c_;b_])]) in
  let habc = T.Fun(h, [a_;T.Fun(f, [b_;c_])]) in
  let fgaab = T.Fun(f, [ga;T.Fun(f, [a_;b_])]) in
  let faby = T.Fun(f, [a_;T.Fun(f, [b_;y])]) in
  let fxy = T.Fun(f, [x;y]) in
  let fxyxy = T.Fun(f, [fxy;fxy]) in
  let t1 = T.Fun(f, [fxy; y]) in
  let t2 = T.Fun(f, [gu; faaa]) in
  let t3 = T.Fun(f, [habc; fxy]) in
  (* Abelian group *)
  let g0 = R.of_terms (T.Fun(f, [x; gx])) zero in
  let g1 = R.of_terms (T.Fun(f, [x; zero])) x in
  let g2 = R.of_terms (T.Fun(g, [fxy])) (T.Fun(f, [gx; T.Fun(g, [y])])) in
  let g3 = R.of_terms (T.Fun(g, [zero])) zero in
  let g4 = R.of_terms (T.Fun(g, [gx])) x in
  let group = [g0; g1; g2; g3; g4] in
  (* commutative ring *)
  let r1 = R.of_terms (T.Fun(m, [fxy; z]))
                      (T.Fun(f, [T.Fun(m, [x;z]); T.Fun(m, [y;z])])) in
  let r2 = R.of_terms (T.Fun(m, [one; x])) x in
  let r3 = R.of_terms (T.Fun(m, [zero; x])) zero in
  let r4 = R.of_terms (T.Fun(m, [gx; y])) (T.Fun(g, [T.Fun(m, [x;y])])) in
  let ring = [r1; r2; r3; r4] @ group in
  (* distributive lattice *)
  let j,sigma = Sig.create_fun 2 "j" sigma in
  let sigma = Sig.add_theory j Theory.AC sigma in
  let l1 = R.of_terms (T.Fun(m, [T.Fun(j, [x;y]); z]))
                      (T.Fun(j, [T.Fun(m, [x;z]); T.Fun(m, [y;z])])) in
  let l2 = R.of_terms (T.Fun(j, [T.Fun(m, [x; y]); x])) x in
  let l3 = R.of_terms (T.Fun(m, [x; x])) x in
  let l4 = R.of_terms (T.Fun(j, [x; x])) x in
  let dlattice = Trs.of_list [l1; l2; l3; l4] in
  (* testing acpos *)
  let app_string s (p, t) =
   assert (try let _ = T.subterm p t in true with _ -> false);
   T.to_stringm t >>= fun ts ->
   return (s ^ "\n (" ^ ts ^ ", " ^ (P.to_string p) ^ ")")
  in
  let show n ps =
   Monad.foldl app_string "" ps >>= fun s -> 
   Format.printf "%s\n" s;
   assert (List.length ps == n);
   return ()
  in 
  let check_acpos t n = Either.right (Monad.run sigma (
   T.to_stringm t >>= fun ts ->
   Format.printf "AC positions in %s:" ts;
   ac_pos t >>= (show n)))
  in
  check_acpos faby 7;
  check_acpos fgaab 8;
  check_acpos t1 5;
  check_acpos t2 8;
  check_acpos t3 11;
  check_acpos fxy 3;
  check_acpos (R.lhs r1) 5;
  (* testing rewrite *)
  let reducts t rule res = (
   T.to_stringm t >>= fun ts ->
   R.to_stringm rule >>= fun rs ->
   direct_reducts t (Trs.of_list [rule]) >>= fun rds ->
   let app s u = T.to_stringm u >>= fun v -> return (s^"\n "^v) in
   Monad.foldl app "" rds >>= fun s ->
   Format.printf "Rewriting %s with %s yields %s\n%!" ts rs s;
   Monad.for_all (fun t -> Monad.exists (acequivalent t) rds) res >>= fun eq ->
   assert (List.length res = (List.length rds) && eq);
   return ())
  in
  let check_reducts t rl res =
   Either.right (Monad.run sigma (reducts t rl res))
  in
  check_reducts fxy (R.of_terms fxy d_) [d_];
  check_reducts fabc (R.of_terms fxy d_)
   [d_; T.Fun(f, [a_;d_]); T.Fun(f, [b_;d_]); T.Fun(f, [c_;d_])];
  check_reducts t2 (R.of_terms gu d_) [T.Fun(f, [faaa; d_])];
  check_reducts fabc (R.of_terms gu d_) [];
  check_reducts fabc (R.of_terms fxx gu) [];
  check_reducts faaa (R.of_terms fxx x) [faa];
  check_reducts faaaa (R.of_terms fxx x) [faa; faaa];
  check_reducts gfaa (R.of_terms fxx x) [T.Fun(g, [a_])];
  check_reducts fxyxy (R.of_terms fxx x)
   [fxy; T.Fun(f, [x; fxy]); T.Fun(f, [y; fxy])];
  (* testing overlaps and CPs *)
  let cps_trs trs res =
   Trs.to_stringm trs >>= fun trss ->
   cps trs >>= fun cps ->
   let app s uv =
    Monad.project T.to_stringm uv >>= fun (u,v) -> return (s^"\n "^u^" = "^v)
   in
   Monad.foldl app "" cps >>= fun s ->
   (let pair_aceq (s,t) uv =
    Monad.cross (acequivalent s, acequivalent t) uv >>= fun (a,b) ->
    return (b && a)
   in
   Monad.filter (Monad.liftM not @.@
                 (flip (uncurry are_joinable) trs)) cps >>= fun cps ->
   Monad.foldl app "" cps >>= fun s ->
   Format.printf "AC CPs for {%s} are %s\n%!" trss s;
   Monad.for_all (fun cp -> Monad.exists (pair_aceq cp) cps) res >>= fun eq ->
   assert (List.length res == (List.length cps) && eq);
   return ())
  in
  let check_cps trs res = Either.right (Monad.run sigma (cps_trs trs res)) in
  let rl1 = R.of_terms ga a_ in
  let rl2 = (R.of_terms fxx x) in
  let rl3 = R.of_terms fab e_ in
  let rl4 = R.of_terms fac d_ in
  let rl5 = R.of_terms (T.Fun(f, [x; one])) x in
  let rl6 = R.of_terms fxy x in
  let rl7 = R.of_terms (T.Fun(f, [x;a_])) b_ in
  let rl8 = R.of_terms (T.Fun(f, [x;a_])) (T.Fun(m, [x;a_])) in
  let fce, fbd = T.Fun(f, [c_;e_]), T.Fun(f, [b_;d_]) in
  check_cps (Trs.of_list [rl3; rl4]) [fce,fbd; fbd, fce];
  let wcr trs res = (
   Trs.to_stringm trs >>= fun trss ->
   is_wcr trs >>= fun b ->
   Format.printf "TRS {%s} is AC wcr: %d\n%!" trss (if b then 1 else 0);
   assert (b == res);
   return ())
  in
  let check_wcr trs res = Either.right (Monad.run sigma (wcr trs res)) in
  check_wcr (Trs.of_list [rl1]) true;
  check_wcr (Trs.of_list [rl2]) true;
  check_wcr (Trs.of_list [rl3; rl4]) false;
  check_wcr (Trs.of_list [rl2; rl3; rl4]) false;
  check_wcr (Trs.of_list [rl5; g1]) false;
  check_wcr (Trs.of_list [rl6]) false;
  check_wcr (Trs.of_list [rl7]) false;
  check_wcr (Trs.of_list [rl8]) false;
  check_wcr (Trs.of_list group) true;
  check_wcr (Trs.of_list ring) true;
  check_wcr (Trs.of_list dlattice) true
;;
end
