(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** MODULES ******************************************************************)
module Fun = Function;;
module M = Monad;;
module Var = Variable;;

(*** EXCEPTIONS ***************************************************************)
exception Not_semi_unifiable;;
exception Not_unifiable;;
exception Not_matchable;;

(*** OPENS ********************************************************************)
open Prelude;;
open Util;;

(*** MODULES ******************************************************************)
module Make (L : LABEL) = struct
 (*** MODULES *****************************************************************)
 module M = M.Make (L);;
 module S = Substitution.Make (L);;
 module Sig = Signature.Make (L);;
 module Term = Term.Make (L);;

 module TermMap =
 struct
   module M = Map.Partial (Fun);;
   type t = Term.t list M.t

   let find f m = try M.find f m with Not_found -> [];;
   let update f g m = M.add f (g (find f m)) m;;
   let rec of_terms = function
     | (Term.Fun (f, _) as t) :: ts ->
       of_terms ts
       |> update f (List.insert t)
     | Term.Var _ :: _ -> failwith "variable left-hand side"
     | [] -> M.empty
   ;;
 end
 
 (*** TYPES *******************************************************************)
 type 'a m = 'a M.t;;
 type substitution = S.t;;
 type term = Term.t = Var of Var.t | Fun of Fun.t * term list;;
 
 (*** EXCEPTIONS **************************************************************)
 exception Not_semi_unifiable = Not_semi_unifiable;;
 exception Not_unifiable = Not_unifiable;;
 exception Not_matchable = Not_matchable;;
 
 (*** FUNCTIONS ***************************************************************)
 let (>>=) = M.(>>=);;
 
 (* Renaming *)
 let renaming =
  let rec renaming s = function
   | Var x ->
    if S.mem x s then M.return s
    else M.fresh_var >>= fun y -> M.return (S.add x (Var y) s)
   | Fun (f,ts) -> M.foldl renaming s ts
  in
  M.foldl renaming S.empty
 ;;
 
 (* Unification *)
 (* TODO *)
 (* let semi_unify_problem p = raise Not_semi_unifiable;; *)
 
 (* t semi_unify u v = semi_unify_problem [(u,v)];; *)
 
 let unify_problem =
  let rec unify s = function
   | [] -> s
   | (u,v) :: p when u = v -> unify s p
   | (Var x as u,v) :: p | (v,(Var x as u)) :: p ->
    if not (Term.is_subterm u v) && S.apply x v s = v then
     let t = S.singleton x v and apply = S.apply_term in
     unify (S.compose apply s t) (List.rev_map (Pair.map (apply t)) p)
    else raise Not_unifiable
   | (Fun (f,us),Fun (g,vs)) :: p ->
    if f = g then unify s (List.rev_append (List.rev_zip us vs) p)
    else raise Not_unifiable
  in
  unify S.empty
 ;;
 
 let unify u v = unify_problem [(u,v)];;
 
 (*
 let are_semi_unifiable u =
  catch Not_semi_unifiable (const true @.@ semi_unify u) false
 ;;
 *)
 
 let are_unifiable u = catch Not_unifiable (const true @.@ unify u) false;;
 
 (* Matching *)
 (* general matching *)
 let match_problem =
  let rec solve s = function
   | [] -> s
   | (u,Var x) :: p ->
    (try solve (S.add x u s) p with S.Inconsistent -> raise Not_matchable)
   | (Fun (f,us),Fun (g,vs)) :: p ->
    if f = g then solve s (List.rev_append (List.rev_zip us vs) p)
    else raise Not_matchable
   | _ -> raise Not_matchable
  in
  solve S.empty
 ;;
 
 let match_term u v = match_problem [(u,v)];;
 let matches u = catch Not_matchable (const true @.@ match_term u) false;;
 let is_variant u v = matches u v && matches v u
 let subsumes = flip matches;;
 let contains u v = List.exists (subsumes v) (Term.subterms u);;
 
(* ground matching *)
let rec simplify = function
  | [] -> Some []
  | (Var _, _) :: ps -> simplify ps
  | (Fun (f, us), Fun (g, vs)) :: ps ->
    if f = g then simplify (List.zip us vs @ ps) else None
  | (t, Var x) :: ps -> Option.map (List.cons (t, x)) (simplify ps)
;;

let rec merge s t = match (s, t) with
  | Var _, u | u, Var _ -> Some u
  | Fun (f, us), Fun (g, vs) ->
    if f = g then
      List.foldr2 (fun u v -> Option.flat_map (fun ws ->
        merge u v |> Option.map (fun x -> x :: ws))) (Some []) us vs
      |> Option.map (Term.make_fun f)
    else None
;;

let rec extend x ps u = match ps with
  | [] -> Some []
  | (v, y) :: ps ->
    if x = y then Option.flat_map (extend x ps) (merge u v)
    else Option.map (List.cons (v, y)) (extend x ps u)
;;

let rec solve = function
 | [] -> true
 | (u, x) :: ps -> Option.fold solve false (extend x ps u)
;;

let ground_matches_list = Option.fold solve false @.@ simplify;;

let ground_matches u v = ground_matches_list [(u, v)];;

let tcap' s m =
  let rec tcap = function
    | Term.Fun (f, ts) ->
      let t = Term.Fun (f, List.map tcap ts) in
      if List.exists (ground_matches t) (TermMap.find f m) then s else t
    | x -> x
  in
  tcap
;;

end
